
###################################################
########## Table 1: Critical Values ###############
###################################################

start <- Sys.time()
cv.SN <- SN.W(AnzSim = 10000, AnzGitter = 100000, t0=0.2, alpha.q = c(0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 0.99))$q
print(Sys.time() - start)

SN.W <- function(AnzSim, AnzGitter, t0, alpha.q){  # AnzSim = Anzahl Wiener-Pfade, AnzGitter = Diskretisierungspunkte der Wiener-Pfade, Quantile = Vektor der zu berechnenden Quantile
  ZV     <- numeric(AnzSim)                        # An Gitterpunkten (in t) wird (B(t)-tB(1))^{2} ausgewertet
  for(j in 1 : AnzSim){
    B <- rnorm(AnzGitter)                   # fill vector with N(0,1)-r.v.s
    B <- 1 / sqrt(AnzGitter) * cumsum(B)    # this is the std. Brownian motion
    t <- floor(AnzGitter * t0) : AnzGitter  # range of t
    ZV[j] <- B[AnzGitter]^2 / ( 1/AnzGitter * sum( (B[t] - t/ AnzGitter * B[AnzGitter])^2 ) )                             # Eine Realisation wird gespeichert
  }
  return(list(q = quantile(ZV, probs = alpha.q), obs = ZV))       # quantile-Funktion liefert die gewuenschten Quantile    
}  


######################################################
########## Table 2: Simulation Results ###############
######################################################


library(snowfall)     # Parallel computing

sfInit(parallel = TRUE, cpus = 3, type = "SOCK")
#sfInit(parallel = FALSE)   # useful for trouble-shooting
sfLibrary(fGarch)
sfLibrary(sn)
sfLibrary(zoo)
sfLibrary(actuar)
sfExport("CI.90.optim.k", "coverage.optim.k", "garchsim_t", "garchsim_alt", "q.st")


wrapper <- function(d, alpha, n, v){
  CI.90.optim.k(d, AnzSim = 10000, alpha, n, t0 = 0.2, cv.SN = 55.65, cv.NA = 1.96, v = 10)
}

start <- Sys.time()

res.05  <- sfLapply(1:6, wrapper, alpha = 0.005,  n = 1000)
res.05[[1]]$res    # [Spi16, Table 1 - Case 6]
res.05[[2]]$res    # [BJW15] values for GARCH(1,1) model
res.05[[3]]$res    # [BJW15] values for AR(1)-GARCH(1,1) model
res.05[[4]]$res    # t(3)
res.05[[5]]$res    # Burr(1.5)
res.05[[6]]$res    # Burr(1)

res.05[[1]]$ind    # [Spi16, Table 1 - Case 6]
res.05[[2]]$ind    # [BJW15] values for GARCH(1,1) model
res.05[[3]]$ind    # [BJW15] values for AR(1)-GARCH(1,1) model

res.1  <- sfLapply(1:6, wrapper, alpha = 0.01,  n = 1000)
res.1[[1]]$res    # [Spi16, Table 1 - Case 6]
res.1[[2]]$res    # [BJW15] values for GARCH(1,1) model
res.1[[3]]$res    # [BJW15] values for AR(1)-GARCH(1,1) model
res.1[[4]]$res    # t(3)
res.1[[5]]$res    # Burr(1.5)
res.1[[6]]$res    # Burr(1)

res.1[[1]]$ind    # [Spi16, Table 1 - Case 6]
res.1[[2]]$ind    # [BJW15] values for GARCH(1,1) model
res.1[[3]]$ind    # [BJW15] values for AR(1)-GARCH(1,1) model


res.25  <- sfLapply(1:6, wrapper, alpha = 0.025,  n = 1000)
res.25[[1]]$res    # [Spi16, Table 1 - Case 6]
res.25[[2]]$res    # [BJW15] values for GARCH(1,1) model
res.25[[3]]$res    # [BJW15] values for AR(1)-GARCH(1,1) model
res.25[[4]]$res    # t(3)
res.25[[5]]$res    # Burr(1.5)
res.25[[6]]$res    # Burr(1)

res.25[[1]]$ind    # [Spi16, Table 1 - Case 6]
res.25[[2]]$ind    # [BJW15] values for GARCH(1,1) model
res.25[[3]]$ind    # [BJW15] values for AR(1)-GARCH(1,1) model

print(Sys.time() - start)

sfStop()


##################################################################
########## Figure 1: PP plots for GARCH(1,1) model ###############
##################################################################


real.V <- SN.W(AnzSim = 100000, AnzGitter = 100000, t0=0.2, alpha.q = c(0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 0.99))$obs

split.screen(figs=c(2, 2))

screen(1)                            
par(mar = c(4.5, 4, 1.5, 2))
pp.plot(rnorm(1000000), res.1[[2]]$hist[3, ])  # Hist.H.ES.NA

screen(2)                            
par(mar = c(4.5, 4, 1.5, 2))
pp.plot(rnorm(1000000), res.1[[2]]$hist[7, ])  # Hist.MR.ES.NA

screen(3)                            
par(mar = c(5.5, 4, 0.5, 2))
pp.plot(real.V, res.1[[2]]$hist[4, ])             # Hist.H.ES.SN
mtext("(a) Hill estimator", side = 1, line=4)

screen(4)                            
par(mar = c(5.5, 4, 0.5, 2))
pp.plot(real.V, res.1.EQ.2nalpha[[2]]$hist[8, ])  # Hist.MR.ES.SN
mtext("(a) MR estimator", side = 1, line=4)

close.screen(all = TRUE)


pp.plot <- function(x, y){
  ecdf.y <- ecdf(y)
  ecdf.x <- ecdf(x)
  x.q <- quantile(x, probs=seq(0.01, 0.99, 0.01), type=8) # The values at which we want to evaluate the empirical CDF
  p.x <- ecdf.x(x.q) 
  p.y <- ecdf.y(x.q)
  plot(p.x, p.y, xlim=c(0, 1), ylim=c(0, 1), xlab="", ylab = "", xaxs="i", yaxs="i")
  mtext("Theoretical Probabilities", side = 1, line=2.5)
  mtext("Sample Probabilities", side = 2, line=2.5)
  
  abline(a=0, b=1)
}


#########################################################
########## Functions to replicate Table 2 ###############
#########################################################


# Simulations with optimal k

CI.90.optim.k <- function(d, AnzSim, alpha, n, t0, cv.SN, cv.NA, v){
  hist.names <- list( c("Hist.H.VaR.NA", "Hist.H.VaR.SN", "Hist.H.ES.NA", "Hist.H.ES.SN", 
                        "Hist.MR.VaR.NA", "Hist.MR.VaR.SN", "Hist.MR.ES.NA", "Hist.MR.ES.SN") )
  histogram  <- matrix(NA, nrow = 8, ncol = AnzSim, dimnames = hist.names)
  ind.names  <- list( c("H", "MR"), c("Percentage > 0.9"))
  indicator  <- matrix(0, nrow=2, ncol=1)
  res.H <- res.MR <- res.CV1 <- matrix(0, nrow = 2, ncol = 7)
  
  for(i in 1:AnzSim){
    if(d == 1){  # [Spi16, Table 1 - Case 6]
      cov    <- coverage.optim.k(ARMA.coef = c(0, 0), GARCH.coef = c(0.95 * 20^2 / 252, 0.1, 0.85), alpha, nu = 3, xi = 5, innov = "t", n, t0, cv.SN, cv.NA, v)  
    }
    if(d == 2){  # [BJW15] values for ARMA(0,0)-GARCH(1,1) model
      cov    <- coverage.optim.k(ARMA.coef = c(0, 0.0), GARCH.coef = c(3.244281e-06, 0.0349, 0.9373), alpha, nu = 4.2, xi = 0, innov = "t", n, t0, cv.SN, cv.NA, v)
    }
    if(d == 3){  # [BJW15] values for ARMA(1,0)-GARCH(1,1) model
      cov    <- coverage.optim.k(ARMA.coef = c(0.2714, 0.0), GARCH.coef = c(3.394708e-06, 0.1407, 0.7914), alpha, nu = 5.3, xi = 0.8531, innov = "t", n, t0, cv.SN, cv.NA, v)
    }
    ############## Appendix B ##############
    if(d == 4){  
      cov    <- coverage.optim.k(ARMA.coef = c(0, 0), GARCH.coef = c(0.95 * 20^2 / 252, 0.1, 0.85), alpha, nu = 3, xi = 0, innov = "t", n, t0, cv.SN, cv.NA, v)  
    }
    if(d == 5){  
      cov    <- coverage.optim.k(ARMA.coef = c(0, 0), GARCH.coef = c(0.95 * 20^2 / 252, 0.1, 0.85), alpha, nu = 3, xi = 0, innov = "Burr1", n, t0, cv.SN, cv.NA, v)  
    }
    if(d == 6){  
      cov    <- coverage.optim.k(ARMA.coef = c(0, 0), GARCH.coef = c(0.95 * 20^2 / 252, 0.1, 0.85), alpha, nu = 3, xi = 0, innov = "Burr2", n, t0, cv.SN, cv.NA, v)  
    }
    ############## Appendix B ##############
    
    indicator       <- indicator + cov$ind
    histogram[ , i] <- cov$hist
    cov.H           <- cov$res.H
    cov.MR          <- cov$res.MR
    cov.CV1         <- cov$res.CV1
    res.H           <- res.H   + cov.H
    res.MR          <- res.MR  + cov.MR 
    res.CV1         <- res.CV1 + cov.CV1 
  }
  
  indicator <- 100 * indicator / AnzSim   # indicator is in %
  
  res.H     <- res.H  / AnzSim
  res.MR    <- res.MR / AnzSim
  res.CV1   <- res.CV1/ AnzSim
  
  # Save results for CVaR and CES
  results           <- rbind(res.H, res.MR, res.CV1)
  results[ , 3]     <- sqrt( results[ , 3] )    # take root for RMSE
  dimnames(results) <- list( c("H", "CES H", "MR", "CES MR", "CV1", "CES CV1"), 
                             c("k*", "Bias", "RMSE", "Coverage.NA", "Coverage.SN", "IL.NA", "IL.SN") )
  
  return(list(res = results, hist = histogram, ind = indicator))
}

###########################

# Coverage probabilities with optimal k chosen according to [Dea16]

coverage.optim.k <- function(ARMA.coef, GARCH.coef, alpha, nu, xi, innov, n, t0, cv.SN, cv.NA, v){
  H.ind <- MR.ind <- 0    # =1, if H >= 0.9, MR >= 0.9
  
  cond=TRUE
  while (cond==TRUE){
    # 1. Simulate ARMA(1,1)-GARCH(1,1) model with skewed-t innovations 
    if(identical(innov, "t")){
      X <- garchsim_t(ARMA.coef, GARCH.coef, alpha, nu, xi, n, ntrans = 1000)   # X[[1]] contains time series, X[[2]] true CVaR  
    }
    else{
      X <- garchsim_alt(ARMA.coef, GARCH.coef, alpha, innov, n, ntrans = 1000)
    }
    
    # 2. Save true CVaR 
    CES.true  <- X[[3]]
    CVaR.true <- X[[2]]    
    X         <- X[[1]]   # for brevity
    
    # 3. Calculate estimated CVaR based on residuals U^
    if( identical(ARMA.coef, c(0, 0)) ){   # pure GARCH(1,1) case
      if(nu <= 4){   # if innovations have infinite 4th moments, use Laplace QMLE ...
        outp <- try(garchFit(formula = ~garch(1,1), data = X, init.rec = "mci", cond.dist = "ged",
                             shape = 1, include.mean = FALSE, include.shape = FALSE,
                             algorithm = "lbfgsb+nm", trace = FALSE) , silent = TRUE)
      }
      else{          # ... else use standard Gaussian QMLE
        outp <- try(garchFit(formula = ~garch(1,1), data = X, init.rec = "mci", cond.dist = "QMLE",
                             include.mean = FALSE, algorithm = "lbfgsb+nm", trace = FALSE) , silent = TRUE)
      }
      
      garch.coef = try(outp@fit$matcoef, silent=TRUE)
      
      omega1 = try(garch.coef[1],silent=TRUE)
      alpha1 = try(garch.coef[2],silent=TRUE)
      beta1  = try(garch.coef[3],silent=TRUE)
      
      if(is.na(omega1)){omega1=1}
      if(is.na(alpha1)){alpha1=1}
      if(is.na(beta1)){beta1=1}
      
      if(is.character(outp)){cond=TRUE}
      else{
        cond=((beta1>=1)|(omega1<0) | (alpha1<0) |(beta1<0)|((alpha1+beta1)>=1))
      }
      
      theta.hat <- c(0, 0, omega1, alpha1, beta1)   # add 0 for the ARMA parameters
    }
    else if( identical(ARMA.coef[2], 0) ){    # AR(1)-GARCH(1,1) case
      outp <- try(garchFit(formula = ~arma(1,0)+garch(1,1), data = X, init.rec = "mci", cond.dist = "QMLE",
                           include.mean = FALSE, algorithm = "lbfgsb+nm", trace = FALSE) , silent = TRUE)
      garch.coef = try(outp@fit$matcoef,silent=TRUE)
      
      phi1   = try(garch.coef[1],silent=TRUE)
      omega1 = try(garch.coef[2],silent=TRUE)
      alpha1 = try(garch.coef[3],silent=TRUE)
      beta1  = try(garch.coef[4],silent=TRUE)
      
      if(is.na(phi1)){phi1=1}
      if(is.na(omega1)){omega1=1}
      if(is.na(alpha1)){alpha1=1}
      if(is.na(beta1)){beta1=1}
      
      if(is.character(outp)){cond=TRUE}
      else{
        cond=((abs(phi1)>=1) | (beta1>=1)|(omega1<0) | (alpha1<0) |(beta1<0)|((alpha1+beta1)>=1))
      }
      theta.hat <- c(phi1, 0, omega1, alpha1, beta1)   # add 0 for the ARMA parameters
    }
    else{                                  # ARMA(1,1)-GARCH(1,1) case
      outp <- try(garchFit(formula = ~arma(1,1)+garch(1,1), data = X, init.rec = "mci", cond.dist = "QMLE",
                           include.mean = FALSE, algorithm = "lbfgsb+nm", trace = FALSE) , silent = TRUE)
      garch.coef=try(outp@fit$matcoef,silent=TRUE)
      
      phi1    = try(garch.coef[1],silent=TRUE)
      theta1  = try(garch.coef[2],silent=TRUE)
      omega1  = try(garch.coef[3],silent=TRUE)
      alpha1  = try(garch.coef[4],silent=TRUE)
      beta1   = try(garch.coef[5],silent=TRUE)
      
      if(is.na(phi1)){phi1=1}
      if(is.na(theta1)){theta1=1}
      if(is.na(omega1)){omega1=1}
      if(is.na(alpha1)){alpha1=1}
      if(is.na(beta1)){beta1=1}
      
      if(is.character(outp)){cond=TRUE}
      else{
        cond=((abs(phi1)>=1)|(abs(theta1)>=1) | (beta1>=1)|(omega1<0) | (alpha1<0) |(beta1<0)|((alpha1+beta1)>=1))
      }
      theta.hat <- c(phi1, -theta1, omega1, alpha1, beta1)
    }
  }
  
  eps.hat <- c(X[1], coredata(filter(X[2:n] - theta.hat[1] * X[1:(n-1)],
                                     filter = theta.hat[2], method = "recursive", init = X[1])))
  sigma.sq.hat <- c(theta.hat[3], coredata(filter(theta.hat[3] + theta.hat[4] * eps.hat[1:(n-1)]^2,
                                                  filter = theta.hat[5], method = "recursive", init = theta.hat[3])))
  
  U.hat <- -eps.hat / sqrt(sigma.sq.hat)   # [KL16, p. 243]
  U.hat <- c(rep.int(NA, nu-1), U.hat[v:length(U.hat)])    # offset initialization of variance equation
  
  # 3.1 Compute optimal k* using [Dea16+, p. 16] (Tail Index Estimation: Quantile Driven Threshold Selection)
  sort.U <- sort.int(U.hat, decreasing = TRUE, method = "quick")
  
  k.min  <- 50      # minimal value of k
  k.max  <- 200     # maximal value of k
  
  # For CV-Estimator
  Rho1   <- 1
  lo_pre <- -log(1 : (k.max + 1))
  x1     <- (0 : (k.max+1))^(Rho1+1)
  
  sup.H <- sup.MR <- sup.CV1 <- rep.int(NA, k.max)
  
  for(k in k.min : k.max){
    
    # 3.1.1 For CVaR estimation
    H.k    <- 1/k * sum(log( sort.U[1 : (k+1)] / sort.U[(k+1)] ))     # Hill estimate based on (k+1) largest order statistics
    M2     <- 1/k * sum(log( sort.U[1 : (k+1)] / sort.U[(k+1)] )^2)
    MR.k   <- 0.5 * M2 / H.k                                          # MR   estimate based on (k+1) largest order statistics

    # CV-estimate
    lo    <- lo_pre[1 : (k+1)] + log((k+1)+1)
    int1  <- (Rho1+1)/Rho1 * 1/(k+1) * (1 - 1/(k+1)^Rho1 * (x1[2 : (k+2)] - x1[1 : (k+1)]))
    CV1.k  <- sum(int1 * log(sort.U[1 : (k+1)])) / sum(int1 * lo)
    
    sup.H[k]   <- max( abs(sort.U[2 : (k.max+1)] - (k/(1 : k.max))^( H.k ) * sort.U[(k+1)]) )
    sup.MR[k]  <- max( abs(sort.U[2 : (k.max+1)] - (k/(1 : k.max))^( MR.k ) * sort.U[(k+1)]) )
    sup.CV1[k]  <- max( abs(sort.U[2 : (k.max+1)] - (k/(1 : k.max))^( CV1.k ) * sort.U[(k+1)]) )
  }
  k.star.H   <- which.min(sup.H)
  k.star.MR  <- which.min(sup.MR)
  k.star.CV1  <- which.min(sup.CV1)

  
  # 3.2 Calculate sequential VaR and ES estimates using k.star
  m <- n - nu + 1
  H   <- H.EQ   <- H.ES   <- rep.int(NA, n)
  MR  <- MR.EQ  <- MR.ES  <- rep.int(NA, n)
  CV1  <- CV1.EQ  <- CV1.ES  <- rep.int(NA, n)
  # For CV-estimator
  lo_pre <- -log(1 : (k.star.CV1 + 1))
  x1     <- (0 : (k.star.CV1+1))^(Rho1+1)

  
  Y <- sort.int(U.hat[1 : floor(n * t0 - 1)], decreasing = TRUE, method = "quick")
  for(i in floor(n * t0) : n){
    Y   <- sort.int(c(Y, U.hat[i]), decreasing = TRUE, method = "quick")      # sort X_1,...,X_[i] in descending order
    
    # Hill estimator ...
    # ... for CVaR 
    H[i]    <- 1/(k.star.H*(i/n)) * sum(log(Y[1 : (floor(k.star.H*(i/n))+1)] / Y[floor(k.star.H*(i/n))+1])) # Hill estimator
    H.EQ[i] <- -Y[floor(k.star.H*(i/n))+1] * ((n*alpha) / k.star.H)^(-H[i])                                 # Hill EQ estimator
    
    # ... and CES
    H.help <- H[i]
    if(H.help > 0.9){
      H.help <- min(0.9, H.help)
      H.ind  <- 1
    }
    H.ES[i]   <- H.EQ[i] / (1 - H.help)
    
    # MR estimator ...
    # ... for CVaR
    M1       <- 1/(k.star.MR*(i/n)) * sum(log(Y[1 : (floor(k.star.MR*(i/n))+1)] / Y[floor(k.star.MR*(i/n))+1])) # Hill estimator
    M2       <- 1/(k.star.MR*(i/n)) * sum(log(Y[1 : (floor(k.star.MR*(i/n))+1)] / Y[floor(k.star.MR*(i/n)) + 1])^2)
    MR[i]    <- 0.5 * M2 / M1                                              # MR estimator
    MR.EQ[i] <- -Y[floor(k.star.MR*(i/n))+1] * ((n*alpha) / k.star.MR)^(-MR[i])   # MR EQ estimator
    
    # ... and CES
    MR.help <- MR[i]
    if(MR.help > 0.9){
      MR.help <- min(0.9, MR.help)
      MR.ind  <- 1
    }
    MR.ES[i]   <- MR.EQ[i] / (1 - MR.help)
    
    # CV1 estimator ...
    # ... for CVaR
    lo   <- lo_pre[1 : (floor(k.star.CV1*(i/n))+1)] + log((floor(k.star.CV1*(i/n))+1)+1)
    int1  <- (Rho1+1)/Rho1 * 1/(floor(k.star.CV1*(i/n))+1) * (1 - 1/(floor(k.star.CV1*(i/n))+1)^Rho1 * (x1[2 : (floor(k.star.CV1*(i/n))+2)] - x1[1 : (floor(k.star.CV1*(i/n)+1))]))
    CV1[i] <- sum(int1 * log(Y[1 : (floor(k.star.CV1*(i/n))+1)])) / sum(int1 * lo)
    CV1.EQ[i] <- -Y[floor(k.star.CV1*(i/n))+1] * ((n*alpha) / k.star.CV1)^(-CV1[i])   # CV EQ estimator

    # ... and CES
    CV1.help <- CV1[i]
    if(CV1.help > 0.9){
      CV1.help <- min(0.9, CV1.help)
      CV1.ind  <- 1
    }
    CV1.ES[i]   <- CV1.EQ[i] / (1 - CV1.help)
  }
  sigma.hat.n1 <- sqrt( theta.hat[3] + theta.hat[4] * eps.hat[n]^2 + theta.hat[5] * sigma.sq.hat[n] )
  
  CVaR.est.H   <- theta.hat[1] * X[n] + sigma.hat.n1 * H.EQ   - theta.hat[2] * eps.hat[n]   # [HogXX, Eq. (1)]
  CVaR.est.MR  <- theta.hat[1] * X[n] + sigma.hat.n1 * MR.EQ  - theta.hat[2] * eps.hat[n]   # [HogXX, Eq. (1)]
  CVaR.est.CV1 <- theta.hat[1] * X[n] + sigma.hat.n1 * CV1.EQ  - theta.hat[2] * eps.hat[n]   # [HogXX, Eq. (1)]
  
  CES.est.H   <- theta.hat[1] * X[n] + sigma.hat.n1 * H.ES   - theta.hat[2] * eps.hat[n]   # [HogXX, Eq. (1)]
  CES.est.MR  <- theta.hat[1] * X[n] + sigma.hat.n1 * MR.ES  - theta.hat[2] * eps.hat[n]   # [HogXX, Eq. (1)]
  CES.est.CV1 <- theta.hat[1] * X[n] + sigma.hat.n1 * CV1.ES  - theta.hat[2] * eps.hat[n]   # [HogXX, Eq. (1)]
  
  
  # 4. Calculate confidence intervals for CVaR
  CI.lo.MR    <- CI.lo.MR.ES    <- matrix(NA, nrow = 2, ncol = 1)   # 2 rows for normal approximation & self-normalization
  CI.lo.CV1   <- CI.lo.CV1.ES    <- matrix(NA, nrow = 2, ncol = 1)
  CI.lo.H     <- CI.lo.H.ES     <- matrix(NA, nrow = 2, ncol = 1)
  
  CI.up.MR    <- CI.up.MR.ES    <- matrix(NA, nrow = 2, ncol = 1)   # 2 rows for normal approximation & self-normalization
  CI.up.CV1   <- CI.up.CV1.ES    <- matrix(NA, nrow = 2, ncol = 1)
  CI.up.H     <- CI.up.H.ES     <- matrix(NA, nrow = 2, ncol = 1)
  
  coverage.H  <- coverage.H.ES  <- matrix(NA, nrow = 2, ncol = 1)
  coverage.MR <- coverage.MR.ES <- matrix(NA, nrow = 2, ncol = 1)
  coverage.CV1 <- coverage.CV1.ES <- matrix(NA, nrow = 2, ncol = 1)
  
  # 4.1 Normal Approximation ...
  # ... for CVaR
  CI.lo.H[1, ] <- CVaR.est.H[n] * exp( cv.NA * H[n] * log(k.star.H / (n*alpha)) / sqrt(k.star.H) )
  CI.up.H[1, ] <- CVaR.est.H[n] * exp(-cv.NA * H[n] * log(k.star.H / (n*alpha)) / sqrt(k.star.H) )
  
  CI.lo.MR[1, ] <- CVaR.est.MR[n] * exp( cv.NA * sqrt(2) * MR[n] * log(k.star.MR / (n*alpha)) / sqrt(k.star.MR) )
  CI.up.MR[1, ] <- CVaR.est.MR[n] * exp(-cv.NA * sqrt(2) * MR[n] * log(k.star.MR / (n*alpha)) / sqrt(k.star.MR) )
  
  CI.lo.CV1[1, ] <- CVaR.est.CV1[n] * exp( cv.NA * sqrt( (2*Rho1 + 2)/(2*Rho1 + 1) ) * CV1[n] * log(k.star.CV1 / (n*alpha)) / sqrt(k.star.CV1) )
  CI.up.CV1[1, ] <- CVaR.est.CV1[n] * exp(-cv.NA * sqrt( (2*Rho1 + 2)/(2*Rho1 + 1) ) * CV1[n] * log(k.star.CV1 / (n*alpha)) / sqrt(k.star.CV1) )
  
  # ... for CES
  CI.lo.H.ES[1, ] <- CES.est.H[n] * exp( cv.NA * H[n] * log(k.star.H / (n*alpha)) / sqrt(k.star.H) )
  CI.up.H.ES[1, ] <- CES.est.H[n] * exp(-cv.NA * H[n] * log(k.star.H / (n*alpha)) / sqrt(k.star.H) )
  
  CI.lo.MR.ES[1, ] <- CES.est.MR[n] * exp( cv.NA * sqrt(2) * MR[n] * log(k.star.MR / (n*alpha)) / sqrt(k.star.MR) )
  CI.up.MR.ES[1, ] <- CES.est.MR[n] * exp(-cv.NA * sqrt(2) * MR[n] * log(k.star.MR / (n*alpha)) / sqrt(k.star.MR) )
  
  CI.lo.CV1.ES[1, ] <- CES.est.CV1[n] * exp( cv.NA * sqrt( (2*Rho1 + 2)/(2*Rho1 + 1) ) * CV1[n] * log(k.star.CV1 / (n*alpha)) / sqrt(k.star.CV1) )
  CI.up.CV1.ES[1, ] <- CES.est.CV1[n] * exp(-cv.NA * sqrt( (2*Rho1 + 2)/(2*Rho1 + 1) ) * CV1[n] * log(k.star.CV1 / (n*alpha)) / sqrt(k.star.CV1) )

  # 4.2 Self-Normalization ...
  # ... for CVaR
  SN.H  <- 1/n * sum( ( (floor(n*t0) : n)/n * log(CVaR.est.H[floor(n*t0) : n]   / CVaR.est.H[n]) )^2 )
  SN.MR <- 1/n * sum( ( (floor(n*t0) : n)/n * log(CVaR.est.MR[floor(n*t0) : n]  / CVaR.est.MR[n]) )^2 )
  SN.CV1 <- 1/n * sum( ( (floor(n*t0) : n)/n * log(CVaR.est.CV1[floor(n*t0) : n]  / CVaR.est.CV1[n]) )^2 )
  
  CI.lo.H[2, ] <- CVaR.est.H[n] * exp( sqrt(cv.SN * SN.H) )
  CI.up.H[2, ] <- CVaR.est.H[n] * exp(-sqrt(cv.SN * SN.H) )
  
  CI.lo.MR[2, ] <- CVaR.est.MR[n] * exp( sqrt(cv.SN * SN.MR) )
  CI.up.MR[2, ] <- CVaR.est.MR[n] * exp(-sqrt(cv.SN * SN.MR) )
  
  CI.lo.CV1[2, ] <- CVaR.est.CV1[n] * exp( sqrt(cv.SN * SN.CV1) )
  CI.up.CV1[2, ] <- CVaR.est.CV1[n] * exp(-sqrt(cv.SN * SN.CV1) )

  # ... for CES
  SN.H.ES  <- 1/n * sum( ( (floor(n*t0) : n)/n * log(CES.est.H[floor(n*t0) : n]   / CES.est.H[n]) )^2 )
  SN.MR.ES <- 1/n * sum( ( (floor(n*t0) : n)/n * log(CES.est.MR[floor(n*t0) : n]  / CES.est.MR[n]) )^2 )
  SN.CV1.ES <- 1/n * sum( ( (floor(n*t0) : n)/n * log(CES.est.CV1[floor(n*t0) : n]  / CES.est.CV1[n]) )^2 )
  
  CI.lo.H.ES[2, ] <- CES.est.H[n] * exp( sqrt(cv.SN * SN.H.ES) )
  CI.up.H.ES[2, ] <- CES.est.H[n] * exp(-sqrt(cv.SN * SN.H.ES) )
  
  CI.lo.MR.ES[2, ] <- CES.est.MR[n] * exp( sqrt(cv.SN * SN.MR.ES) )
  CI.up.MR.ES[2, ] <- CES.est.MR[n] * exp(-sqrt(cv.SN * SN.MR.ES) )
  
  CI.lo.CV1.ES[2, ] <- CES.est.CV1[n] * exp( sqrt(cv.SN * SN.CV1.ES) )
  CI.up.CV1.ES[2, ] <- CES.est.CV1[n] * exp(-sqrt(cv.SN * SN.CV1.ES) )
 
  # 5. Return if true CVaR / CES covered by c.i. ...
  # ... for CVaR
  coverage.H[1, ]  <- (CI.lo.H[1, ] < CVaR.true) * (CI.up.H[1, ] > CVaR.true)     # = 1, iff  CVaR.true\in(CI.lo.H, CI.up.H)
  coverage.H[2, ]  <- (CI.lo.H[2, ] < CVaR.true) * (CI.up.H[2, ] > CVaR.true)     # = 1, iff  CVaR.true\in(CI.lo.H, CI.up.H)
  
  coverage.MR[1, ] <- (CI.lo.MR[1, ] < CVaR.true) * (CI.up.MR[1, ] > CVaR.true)   # = 1, iff  CVaR.true\in(CI.lo.H, CI.up.H)
  coverage.MR[2, ] <- (CI.lo.MR[2, ] < CVaR.true) * (CI.up.MR[2, ] > CVaR.true)   # = 1, iff  CVaR.true\in(CI.lo.H, CI.up.H)
 
  coverage.CV1[1, ] <- (CI.lo.CV1[1, ] < CVaR.true) * (CI.up.CV1[1, ] > CVaR.true)   # = 1, iff  CVaR.true\in(CI.lo.H, CI.up.H)
  coverage.CV1[2, ] <- (CI.lo.CV1[2, ] < CVaR.true) * (CI.up.CV1[2, ] > CVaR.true)   # = 1, iff  CVaR.true\in(CI.lo.H, CI.up.H)
 
  # ... for CES
  coverage.H.ES[1, ]  <- (CI.lo.H.ES[1, ] < CES.true) * (CI.up.H.ES[1, ] > CES.true)     # = 1, iff  CVaR.true\in(CI.lo.H, CI.up.H)
  coverage.H.ES[2, ]  <- (CI.lo.H.ES[2, ] < CES.true) * (CI.up.H.ES[2, ] > CES.true)     # = 1, iff  CVaR.true\in(CI.lo.H, CI.up.H)
  
  coverage.MR.ES[1, ] <- (CI.lo.MR.ES[1, ] < CES.true) * (CI.up.MR.ES[1, ] > CES.true)   # = 1, iff  CVaR.true\in(CI.lo.H, CI.up.H)
  coverage.MR.ES[2, ] <- (CI.lo.MR.ES[2, ] < CES.true) * (CI.up.MR.ES[2, ] > CES.true)   # = 1, iff  CVaR.true\in(CI.lo.H, CI.up.H)
  
  coverage.CV1.ES[1, ] <- (CI.lo.CV1.ES[1, ] < CES.true) * (CI.up.CV1.ES[1, ] > CES.true)   # = 1, iff  CVaR.true\in(CI.lo.H, CI.up.H)
  coverage.CV1.ES[2, ] <- (CI.lo.CV1.ES[2, ] < CES.true) * (CI.up.CV1.ES[2, ] > CES.true)   # = 1, iff  CVaR.true\in(CI.lo.H, CI.up.H)

  # 6. For histograms
  # ... for CVaR
  Hist.H.VaR.NA  <- 1 / H[n] *              sqrt(k.star.H) / log(k.star.H / (n*alpha)) * log(CVaR.est.H[n] / CVaR.true)
  Hist.MR.VaR.NA <- 1 / (sqrt(2) * MR[n]) * sqrt(k.star.MR) / log(k.star.MR / (n*alpha)) * log(CVaR.est.MR[n] / CVaR.true)
  
  Hist.H.VaR.SN  <- 1 / SN.H  * log(CVaR.est.H[n] / CVaR.true)^2
  Hist.MR.VaR.SN <- 1 / SN.MR * log(CVaR.est.MR[n] / CVaR.true)^2
  
  # ... for CES
  Hist.H.ES.NA  <- 1 / H[n] *              sqrt(k.star.H) / log(k.star.H / (n*alpha)) * log(CES.est.H[n] / CES.true)
  Hist.MR.ES.NA <- 1 / (sqrt(2) * MR[n]) * sqrt(k.star.MR) / log(k.star.MR / (n*alpha)) * log(CES.est.MR[n] / CES.true)
  
  Hist.H.ES.SN  <- 1 / SN.H.ES  * log(CES.est.H[n] / CES.true)^2
  Hist.MR.ES.SN <- 1 / SN.MR.ES * log(CES.est.MR[n] / CES.true)^2
  

  return( list( res.H  = rbind(c(k.star.H, CVaR.est.H[n] - CVaR.true, (CVaR.est.H[n] - CVaR.true)^2, # for CVaR
                            t(coverage.H),    t(CI.up.H - CI.lo.H)),     
                            c(NA, CES.est.H[n] - CES.true, (CES.est.H[n] - CES.true)^2,            # for CES
                            t(coverage.H.ES), t(CI.up.H.ES - CI.lo.H.ES))),     
                res.MR = rbind(c(k.star.MR, CVaR.est.MR[n] - CVaR.true, (CVaR.est.MR[n] - CVaR.true)^2, # for CVaR
                                t(coverage.MR),    t(CI.up.MR - CI.lo.MR)),     
                              c(NA, CES.est.MR[n] - CES.true, (CES.est.MR[n] - CES.true)^2,            # for CES
                              t(coverage.MR.ES), t(CI.up.MR.ES - CI.lo.MR.ES))), 
                res.CV1 = rbind(c(k.star.CV1, CVaR.est.CV1[n] - CVaR.true, (CVaR.est.CV1[n] - CVaR.true)^2, # for CVaR
                                 t(coverage.CV1),    t(CI.up.CV1 - CI.lo.CV1)),     
                               c(NA, CES.est.CV1[n] - CES.true, (CES.est.CV1[n] - CES.true)^2,            # for CES
                                 t(coverage.CV1.ES), t(CI.up.CV1.ES - CI.lo.CV1.ES))), 
                hist = c(Hist.H.VaR.NA, Hist.H.VaR.SN, Hist.H.ES.NA, Hist.H.ES.SN, 
                         Hist.MR.VaR.NA, Hist.MR.VaR.SN, Hist.MR.ES.NA, Hist.MR.ES.SN),
                ind = c(H.ind, MR.ind)) )
}

###########################


# Function to simulate ARMA-GARCH processes with alternative innovation distributions

garchsim_alt <- function(ARMA.coef, GARCH.coef, alpha, innov, n, ntrans){      # GARCH(1,1) coefficients - GARCH.coef = c(omega0, psi0, beta0)
  m <- n + ntrans
  
  # 0. Calculate innovations U
  if(identical(innov, "Burr1")){
    lambda <- 1.5
    tau    <- 2
    sd.wt  <- sqrt(mburr(order=2, shape1 = lambda, shape2 = tau))
    U      <- (2 * rbinom(m, 1, p=0.5) - 1) * rburr(m, shape1 = lambda, shape2 = tau) / sd.wt
    
    # Calculate VaR and ES of U
    q.burr   <- function(x) return( -qburr(1 - 2 * x, shape1 = lambda, shape2 = tau) / sd.wt )
    quant.st <- q.burr(alpha)
    ES.st    <- 1/alpha * integrate( q.burr, lower = 0, upper = alpha )$value
  }
  else if(identical(innov, "Burr2")){
    lambda <- 1
    tau    <- 3
    sd.wt  <- sqrt(mburr(order=2, shape1 = lambda, shape2 = tau))
    U      <- (2 * rbinom(m, 1, p=0.5) - 1) * rburr(m, shape1 = lambda, shape2 = tau) / sd.wt
    
    # Calculate VaR and ES of U
    q.burr   <- function(x) return( -qburr(1 - 2 * x, shape1 = lambda, shape2 = tau) / sd.wt )
    quant.st <- q.burr(alpha)
    ES.st    <- 1/alpha * integrate( q.burr, lower = 0, upper = alpha )$value
  }
  
  # 1. GARCH(1,1) process eps
  eps <- sigma.sq <- numeric(m)                             # intializes vectors of y's and sigma^2's to be m long
  eps[1]      <- rnorm(1)                                   # start value
  sigma.sq[1] <- rnorm(1)^2                                 # start value
  for(i in 2 : m){
    sigma.sq[i] <- GARCH.coef[1] + GARCH.coef[2] * eps[i-1]^2 + GARCH.coef[3] * sigma.sq[i-1]
    eps[i]      <- sqrt(sigma.sq[i]) * U[i]    # Simulate GARCH(1,1) process
  }
  
  # 2. Calculate ARMA(1,1) process X
  X    <- numeric(m)
  X[1] <- rnorm(1)  # random initialization for X
  for(i in 2 : m){
    X[i] <- ARMA.coef[1] * X[i-1] + eps[i] - ARMA.coef[2] * eps[i-1]   # ARMA(1,1)-recursion
  }
  
  # 3. Calcuate true CVaR ...
  sigma.sq.np1 <- GARCH.coef[1] + GARCH.coef[2] * eps[m]^2 + GARCH.coef[3] * sigma.sq[m]
  CVaR.true    <- ARMA.coef[1] * X[m] + sqrt(sigma.sq.np1) * quant.st  - ARMA.coef[2] * eps[m]
  
  # ... and CES
  CES.true     <- ARMA.coef[1] * X[m] + sqrt(sigma.sq.np1) * ES.st - ARMA.coef[2] * eps[m]
  
  return( list(X[(ntrans + 1) : m], CVaR.true, CES.true) )
}


# Function to simulate ARMA-GARCH processes with skewed t innovations

garchsim_t <- function(ARMA.coef, GARCH.coef, alpha, nu, xi, n, ntrans){      # GARCH(1,1) coefficients - GARCH.coef = c(omega0, psi0, beta0)
  m <- n + ntrans
  
  # 1. Calculate GARCH(1,1) innovations eps
  # 1.1 Skewed t-Distribution
  U    <- rst(m, 0, omega = 1, alpha = xi, nu)             # skewed-t_{\nu} distribution 
  E.wt <- xi / sqrt(1 + xi^2) * sqrt(nu / pi) * gamma(0.5 * (nu - 1)) / gamma(0.5 * nu)
  U    <- (U - E.wt) / sqrt( nu / (nu - 2) - E.wt^2 )      # standardized distr. with E=0, Var=1, see [Spi16, Appendix B]
  
  eps <- sigma.sq <- numeric(m)                             # intializes vectors of y's and sigma^2's to be m long
  eps[1]      <- sqrt((nu - 2) / nu) * rt(1, df = nu)       # start value
  sigma.sq[1] <- rnorm(1)^2                                 # start value
  for(i in 2 : m){
    sigma.sq[i] <- GARCH.coef[1] + GARCH.coef[2] * eps[i-1]^2 + GARCH.coef[3] * sigma.sq[i-1]
    eps[i]      <- sqrt(sigma.sq[i]) * U[i]    # Simulate GARCH(1,1) process
  }
  
  # 2. Calculate ARMA(1,1) process X
  X    <- numeric(m)
  X[1] <- rnorm(1)  # random initialization for X
  for(i in 2 : m){
    X[i] <- ARMA.coef[1] * X[i-1] + eps[i] - ARMA.coef[2] * eps[i-1]   # ARMA(1,1)-recursion
  }
  
  # 3. Calcuate true CVaR ...
  sigma.sq.np1 <- GARCH.coef[1] + GARCH.coef[2] * eps[m]^2 + GARCH.coef[3] * sigma.sq[m]
  quant.st     <- ( qst(p = alpha, xi=0, omega = 1, alpha = xi, nu) - E.wt ) / sqrt( nu / (nu - 2) - E.wt^2 )
  CVaR.true    <- ARMA.coef[1] * X[m] + sqrt(sigma.sq.np1) * quant.st  - ARMA.coef[2] * eps[m]
  
  # ... and CES
  quant.st     <- ( qst(p = alpha, xi=0, omega = 1, alpha = xi, nu) - E.wt ) / sqrt( nu / (nu - 2) - E.wt^2 )
  ES.st        <- 1/alpha * integrate( q.st, lower = 0, upper = alpha, xi, nu )$value
  CES.true     <- ARMA.coef[1] * X[m] + sqrt(sigma.sq.np1) * ES.st - ARMA.coef[2] * eps[m]
  
  return( list(X[(ntrans + 1) : m], CVaR.true, CES.true) )
}


# Calculate quantile function of standardized skewed-t distribution
q.st <- function(x, xi, nu){
  E.wt  <- xi / sqrt(1 + xi^2) * sqrt(nu / pi) * gamma(0.5 * (nu - 1)) / gamma(0.5 * nu)
  if(xi == 0){     # for standard symmetric t-distribution
    quant.st <- ( qt(x, df= nu) - E.wt ) / sqrt( nu / (nu - 2) - E.wt^2 )
  }
  else{            # for skewed t-distribution
    quant.st     <- ( qst(x, 0, 1, xi, nu) - E.wt ) / sqrt( nu / (nu - 2) - E.wt^2 )
  }
  return(quant.st)
}



################################################################################
####################### Simulations with wide range of k #######################
################################################################################


#############################################
############### Figures 2 & 3 ###############
#############################################


library(snowfall)     # Parallel computing

sfInit(parallel = TRUE, cpus = 3, type = "SOCK")
#sfInit(parallel = FALSE)   # for trouble-shooting
sfExport("CI.90", "coverage", "garchsim_t", "sf.st")
sfLibrary(fGarch)
sfLibrary(sn)
sfLibrary(zoo)

wrapper <- function(d, alpha, n){
  CI.90(d, AnzSim = 10000, alpha, n)
}

start <- Sys.time()

# Results for Figures 2-4
res.1   <- sfLapply(1:3, wrapper, alpha = 0.01, n = 1000)

print(Sys.time() - start)

# plots for alpha = 0.01
# Figure 2
Spi16.1 <- res.1[[1]]
plot.coverage(Spi16.1[[1]], Spi16.1[[2]], Spi16.1[[3]], Spi16.1[[4]], Spi16.1[[5]], Spi16.1[[6]], 
              Spi16.1[[7]], Spi16.1[[8]], Spi16.1[[9]], Spi16.1[[10]], Spi16.1[[11]], 
              Spi16.1[[12]], Spi16.1[[13]])

# Figure 3
BJW15a.1 <- res.1[[2]]
plot.coverage(BJW15a.1[[1]], BJW15a.1[[2]], BJW15a.1[[3]], BJW15a.1[[4]], BJW15a.1[[5]], 10^3*BJW15a.1[[6]], 
              10^3*BJW15a.1[[7]], 10^3*BJW15a.1[[8]], 10^3*BJW15a.1[[9]], 10^3*BJW15a.1[[10]], 10^3*BJW15a.1[[11]],
              10^3*BJW15a.1[[12]], 10^3*BJW15a.1[[13]])

# Figure 4 (NOT REPORTED IN PAPER)
BJW15b.1 <- res.1[[3]]
plot.coverage(BJW15b.1[[1]], BJW15b.1[[2]], BJW15b.1[[3]], BJW15b.1[[4]], BJW15b.1[[5]], 10^3*BJW15b.1[[6]], 
              10^3*BJW15b.1[[7]], 10^3*BJW15b.1[[8]], 10^3*BJW15b.1[[9]], 10^3*BJW15b.1[[10]], 10^3*BJW15b.1[[11]], 
              10^3*BJW15b.1[[12]], 10^3*BJW15b.1[[13]])

sfStop()


####################################################################
############### Functions to replicate Figures 2 & 3 ############### 
####################################################################


CI.90 <- function(d, AnzSim, alpha, n){
  k.min <- floor(0.05 * n)  # minimal number of upper order statistics used in finding optimal k
  k.max <- floor(0.2 * n)   # maximal number of upper order statistics used in finding optimal k
  k <- k.min : k.max
  cov.H <- cov.MR <- IL.H <- IL.MR <- H.Bias <- H.RMSE <- MR.Bias <- MR.RMSE <- rep.int(0, length(k))
  for(i in 1:AnzSim){
    if(d == 1){  # [Spi16, Table 1 - Case 6]
      cov    <- coverage(ARMA.coef = c(0, 0), GARCH.coef = c(0.95 * 20^2 / 252, 0.1, 0.85), alpha, nu = 3, xi = 5, n, k)
    }
    if(d == 2){  # [BJW15] values for ARMA(0,0)-GARCH(1,1) model
      cov    <- coverage(ARMA.coef = c(0, 0.0), GARCH.coef = c(3.244281e-06, 0.0349, 0.9373), alpha, nu = 4.2, xi = 0, n, k)
    }
    if(d == 3){  # [BJW15] values for ARMA(1,0)-GARCH(1,1) model
      cov    <- coverage(ARMA.coef = c(0.2714, 0.0), GARCH.coef = c(3.394708e-06, 0.1407, 0.7914), alpha, nu = 5.3, xi = 0.8531, n, k)
    }
    cov.H   <- cov.H  + cov[[1]]
    cov.MR  <- cov.MR + cov[[2]]
    IL.H    <- IL.H  + cov[[3]]
    IL.MR   <- IL.MR + cov[[4]]
    H.Bias  <- H.Bias + cov[[5]]
    H.RMSE  <- H.RMSE + cov[[5]]^2
    MR.Bias <- MR.Bias + cov[[6]]
    MR.RMSE <- MR.RMSE + cov[[6]]^2
  }
  coverage.H.NA  <- cov.H[1, ] / AnzSim     # coverage probability for Hill estimator and normal approx.
  coverage.H.SN  <- cov.H[2, ] / AnzSim     # coverage probability for Hill estimator and self-norm.
  coverage.MR.NA <- cov.MR[1, ] / AnzSim    # ...
  coverage.MR.SN <- cov.MR[2, ] / AnzSim
  IL.H.NA  <- IL.H[1, ] / AnzSim            # interval length for Hill estimator and normal approx.
  IL.H.SN  <- IL.H[2, ] / AnzSim            # interval length for Hill estimator and self-norm.
  IL.MR.NA <- IL.MR[1, ] / AnzSim           # ...
  IL.MR.SN <- IL.MR[2, ] / AnzSim
  H.Bias   <- H.Bias  / AnzSim              # Simulation Bias and ...
  MR.Bias  <- MR.Bias / AnzSim             
  H.RMSE   <- sqrt(H.RMSE  / AnzSim)        # ... RMSE
  MR.RMSE  <- sqrt(MR.RMSE / AnzSim)              
  return( list(k, coverage.H.NA, coverage.H.SN, coverage.MR.NA, coverage.MR.SN, 
               IL.H.NA, IL.H.SN, IL.MR.NA, IL.MR.SN, H.Bias, H.RMSE, MR.Bias, MR.RMSE) )
}

###########################

# funtion to estimate CVaR and check if estimate is in 95%-c.i.

coverage <- function(ARMA.coef, GARCH.coef, alpha, nu, xi, n, k, t0 = 0.2, cv.SN = 55.65, cv.NA = 1.96, v=10){
  No.k <- length(k)
  
  cond=TRUE
  while (cond==TRUE){
    # 1. Simulate ARMA(1,1)-GARCH(1,1) model with skewed-t innovations (using fGarch package)
    X <- garchsim_t(ARMA.coef, GARCH.coef, alpha, nu, xi, n, ntrans = 1000)   # X[[1]] contains time series, X[[2]] true CVaR
    
    # 2. Save true CVaR 
    CES.true  <- X[[3]]
    CVaR.true <- X[[2]]    
    X         <- X[[1]]   # for brevity
    
    # 3. Calculate estimated CVaR based on residuals U^
    if( identical(ARMA.coef, c(0, 0)) ){   # pure GARCH(1,1) case
      if(nu <= 4){   # if innovations have infinite 4th moments, use Laplace QMLE ...
        outp <- try(garchFit(formula = ~garch(1,1), data = X, init.rec = "mci", cond.dist = "ged",
                             shape = 1, include.mean = FALSE, include.shape = FALSE,
                             algorithm = "lbfgsb+nm", trace = FALSE) , silent = TRUE)
      }
      else{          # ... else use standard Gaussian QMLE
        outp <- try(garchFit(formula = ~garch(1,1), data = X, init.rec = "mci", cond.dist = "QMLE",
                             include.mean = FALSE, algorithm = "lbfgsb+nm", trace = FALSE) , silent = TRUE)
      }
      garch.coef=try(outp@fit$matcoef,silent=TRUE)
      
      omega1=try(garch.coef[1],silent=TRUE)
      alpha1=try(garch.coef[2],silent=TRUE)
      beta1=try(garch.coef[3],silent=TRUE)
      
      if(is.na(omega1)){omega1=1}
      if(is.na(alpha1)){alpha1=1}
      if(is.na(beta1)){beta1=1}
      
      if(is.character(outp)){cond=TRUE}
      else{
        cond=((beta1>=1)|(omega1<0) | (alpha1<0) |(beta1<0)|((alpha1+beta1)>=1))
      }
      
      theta.hat <- c(0, 0, omega1, alpha1, beta1)   # add 0 for the ARMA parameters
    }
    else if( identical(ARMA.coef[2], 0) ){    # AR(1)-GARCH(1,1) case
      outp <- try(garchFit(formula = ~arma(1,0)+garch(1,1), data = X, init.rec = "mci", cond.dist = "QMLE",
                           include.mean = FALSE, algorithm = "lbfgsb+nm", trace = FALSE) , silent = TRUE)
      garch.coef=try(outp@fit$matcoef,silent=TRUE)
      
      phi1    = try(garch.coef[1],silent=TRUE)
      omega1  = try(garch.coef[2],silent=TRUE)
      alpha1  = try(garch.coef[3],silent=TRUE)
      beta1   = try(garch.coef[4],silent=TRUE)
      
      if(is.na(phi1)){phi1=1}
      if(is.na(omega1)){omega1=1}
      if(is.na(alpha1)){alpha1=1}
      if(is.na(beta1)){beta1=1}
      
      if(is.character(outp)){cond=TRUE}
      else{
        cond=((abs(phi1)>=1) | (beta1>=1)|(omega1<0) | (alpha1<0) |(beta1<0)|((alpha1+beta1)>=1))
      }
      theta.hat <- c(phi1, 0, omega1, alpha1, beta1)   # add 0 for the ARMA parameters
    }
    else{                                  # ARMA(1,1)-GARCH(1,1) case
      outp <- try(garchFit(formula = ~arma(1,1)+garch(1,1), data = X, init.rec = "mci", cond.dist = "QMLE",
                           include.mean = FALSE, algorithm = "lbfgsb+nm", trace = FALSE) , silent = TRUE)
      garch.coef=try(outp@fit$matcoef,silent=TRUE)
      
      phi1    = try(garch.coef[1],silent=TRUE)
      theta1  = try(garch.coef[2],silent=TRUE)
      omega1  = try(garch.coef[3],silent=TRUE)
      alpha1  = try(garch.coef[4],silent=TRUE)
      beta1   = try(garch.coef[5],silent=TRUE)
      
      if(is.na(phi1)){phi1=1}
      if(is.na(theta1)){theta1=1}
      if(is.na(omega1)){omega1=1}
      if(is.na(alpha1)){alpha1=1}
      if(is.na(beta1)){beta1=1}
      
      if(is.character(outp)){cond=TRUE}
      else{
        cond=((abs(phi1)>=1)|(abs(theta1)>=1) | (beta1>=1)|(omega1<0) | (alpha1<0) |(beta1<0)|((alpha1+beta1)>=1))
      }
      theta.hat <- c(phi1, -theta1, omega1, alpha1, beta1)
    }
  }
  
  eps.hat <- c(X[1], coredata(filter(X[2:n]-theta.hat[1]*X[1:(n-1)],
                                     filter = theta.hat[2], method = "recursive", init = X[1])))
  sigma.sq.hat <- c(theta.hat[3], coredata(filter(theta.hat[3] + theta.hat[4] * eps.hat[1:(n-1)]^2,
                                                  filter = theta.hat[5], method = "recursive", init = theta.hat[3])))
  
  U.hat <- -eps.hat / sqrt(sigma.sq.hat)   # [KL16, p. 243]
  U.hat <- c(rep.int(NA, nu-1), U.hat[v:length(U.hat)])    # offset initialization of variance equation

  # 3.1 Calculate sequential EQ estimates
  H <- H.EQ <- H.ES <- MR <- MR.EQ <- MR.ES <- matrix(NA, nrow = n, ncol = No.k)
  
  Y <- sort.int(U.hat[1 : floor(n * t0 - 1)], decreasing = TRUE, method = "quick")
  for(i in floor(n * t0) : n){
    Y   <- sort.int(c(Y, U.hat[i]), decreasing = TRUE, method = "quick")      # sort X_1,...,X_[i] in descending order
    
    for(j in (1 : No.k)){      
      # Hill estimator
      H[i, j]    <- 1/(k[j]*(i/n)) * sum(log(Y[1 : (floor(k[j]*(i/n))+1)] / Y[floor(k[j]*(i/n))+1])) # Hill estimator
      H.EQ[i, j] <- -Y[floor(k[j]*(i/n))+1] * ((n*alpha) / k[j])^(-H[i, j])                            # Hill EQ estimator
      
      H.help <- H[i, j]
      if(H.help > 0.9){
        H.help <- min(0.9, H.help)
      }  
      H.ES[i, j] <- H.EQ[i, j] / (1 - H.help)
      
      # moments ratio estimator
      M1          <- H[i, j]    # just the Hill estimate
      M2          <- 1/(k[j]*(i/n)) * sum(log(Y[1 : (floor(k[j]*(i/n))+1)] / Y[floor(k[j]*(i/n)) + 1])^2)
      MR[i, j]    <- 0.5 * M2/M1                                               # MR estimator
      MR.EQ[i, j] <- -Y[floor(k[j]*(i/n))+1] * ((n*alpha) / k[j])^(-MR[i, j])   # MR EQ estimator
      
      MR.help <- MR[i, j]
      if(MR.help > 0.9){
        MR.help <- min(0.9, MR.help)
      }  
      MR.ES[i, j] <- MR.EQ[i, j] / (1 - MR.help)
    }
  }
  sigma.hat.n1 <- sqrt( theta.hat[3] + theta.hat[4] * eps.hat[n]^2 + theta.hat[5] * sigma.sq.hat[n] )
  CVaR.est.H   <- theta.hat[1] * X[n] + sigma.hat.n1 * H.EQ  - theta.hat[2] * eps.hat[n]  # [HogXX, Eq. (1)]
  CVaR.est.MR  <- theta.hat[1] * X[n] + sigma.hat.n1 * MR.EQ - theta.hat[2] * eps.hat[n]  # [HogXX, Eq. (1)]
  
  CES.est.H    <- theta.hat[1] * X[n] + sigma.hat.n1 * H.ES  - theta.hat[2] * eps.hat[n]  # [HogXX, Eq. (1)]
  CES.est.MR   <- theta.hat[1] * X[n] + sigma.hat.n1 * MR.ES - theta.hat[2] * eps.hat[n]  # [HogXX, Eq. (1)]
  
  # 4. Calculate confidence intervals for CVaR
  rc.names    <- list( c("Normal Approx.", "Self-Norm."), k )
  CI.lo.MR    <- matrix(NA, nrow = 2, ncol = No.k, dimnames = rc.names)   # 2 rows for normal approximation & self-normalization
  CI.lo.H     <- matrix(NA, nrow = 2, ncol = No.k, dimnames = rc.names)
  CI.up.MR    <- matrix(NA, nrow = 2, ncol = No.k, dimnames = rc.names)   # 2 rows for normal approximation & self-normalization
  CI.up.H     <- matrix(NA, nrow = 2, ncol = No.k, dimnames = rc.names)
  coverage.H  <- matrix(NA, nrow = 2, ncol = No.k, dimnames = rc.names)
  coverage.MR <- matrix(NA, nrow = 2, ncol = No.k, dimnames = rc.names)
  
  # 4.1 Normal Approximation
  CI.lo.H[1, ] <- CES.est.H[n, ] * exp( cv.NA * H[n, ] * log(k / (n*alpha)) / sqrt(k) )
  CI.up.H[1, ] <- CES.est.H[n, ] * exp(-cv.NA * H[n, ] * log(k / (n*alpha)) / sqrt(k) )
  
  CI.lo.MR[1, ] <- CES.est.MR[n, ] * exp( cv.NA * sqrt(2) * MR[n, ] * log(k / (n*alpha)) / sqrt(k) )
  CI.up.MR[1, ] <- CES.est.MR[n, ] * exp(-cv.NA * sqrt(2) * MR[n, ] * log(k / (n*alpha)) / sqrt(k) )
  
  # 4.2 Self-Normalization
  SN.H <- SN.MR <- numeric(No.k)
  for(j in 1 : No.k){
    SN.H[j]  <- 1/n * sum( ( (floor(n*t0) : n)/n * log(CES.est.H[floor(n*t0) : n, j] / CES.est.H[n, j]) )^2 )
    SN.MR[j] <- 1/n * sum( ( (floor(n*t0) : n)/n * log(CES.est.MR[floor(n*t0) : n, j] / CES.est.MR[n, j]) )^2 )
  }
  CI.lo.H[2, ] <- CES.est.H[n, ] * exp( sqrt(cv.SN * SN.H) )
  CI.up.H[2, ] <- CES.est.H[n, ] * exp(-sqrt(cv.SN * SN.H) )
  
  CI.lo.MR[2, ] <- CES.est.MR[n, ] * exp( sqrt(cv.SN * SN.MR) )
  CI.up.MR[2, ] <- CES.est.MR[n, ] * exp(-sqrt(cv.SN * SN.MR) )
  
  # 5. Return if true CVaR covered by c.i.
  coverage.H[1, ]  <- (CI.lo.H[1, ]  < CES.true) * (CI.up.H[1, ]  > CES.true)   # = 1, iff  CVaR.true\in(CI.lo.H, CI.up.H)
  coverage.MR[1, ] <- (CI.lo.MR[1, ] < CES.true) * (CI.up.MR[1, ] > CES.true)   # = 1, iff  CVaR.true\in(CI.lo.H, CI.up.H)
  
  coverage.H[2, ]  <- (CI.lo.H[2, ]  < CES.true) * (CI.up.H[2, ]  > CES.true)   # = 1, iff  CVaR.true\in(CI.lo.H, CI.up.H)
  coverage.MR[2, ] <- (CI.lo.MR[2, ] < CES.true) * (CI.up.MR[2, ] > CES.true)   # = 1, iff  CVaR.true\in(CI.lo.H, CI.up.H)

  
  # 6. Return differences to calculate Bias and RMSE
  dev.H  <- CES.est.H[n, ]  - CES.true
  dev.MR <- CES.est.MR[n, ] - CES.true
  
  return( list(coverage.H, coverage.MR, CI.up.H - CI.lo.H, CI.up.MR - CI.lo.MR, dev.H, dev.MR) )
}

###########################

garchsim_t <- function(ARMA.coef, GARCH.coef, alpha, nu, xi, n, ntrans){      # GARCH(1,1) coefficients - GARCH.coef = c(omega0, psi0, beta0)
  m <- n + ntrans
  
  # 1. Calculate GARCH(1,1) innovations eps
  # 1.1 Skewed t-Distribution
  U    <- rst(m, 0, omega = 1, alpha = xi, nu)             # skewed-t_{\nu} distribution 
  E.wt <- xi / sqrt(1 + xi^2) * sqrt(nu / pi) * gamma(0.5 * (nu - 1)) / gamma(0.5 * nu)
  U    <- (U - E.wt) / sqrt( nu / (nu - 2) - E.wt^2 )      # standardized distr. with E=0, Var=1, see [Spi16, Appendix B]
  
  eps <- sigma.sq <- numeric(m)                             # intializes vectors of y's and sigma^2's to be m long
  eps[1]      <- sqrt((nu - 2) / nu) * rt(1, df = nu)       # start value
  sigma.sq[1] <- rnorm(1)^2                                 # start value
  for(i in 2 : m){
    sigma.sq[i] <- GARCH.coef[1] + GARCH.coef[2] * eps[i-1]^2 + GARCH.coef[3] * sigma.sq[i-1]
    eps[i]      <- sqrt(sigma.sq[i]) * U[i]    # Simulate GARCH(1,1) process
  }
  
  # 2. Calculate ARMA(1,1) process X
  X    <- numeric(m)
  X[1] <- rnorm(1)  # random initialization for X
  for(i in 2 : m){
    X[i] <- ARMA.coef[1] * X[i-1] + eps[i] - ARMA.coef[2] * eps[i-1]   # ARMA(1,1)-recursion
  }
  
  # 3. Cacluate true CVaR ...
  sigma.sq.np1 <- GARCH.coef[1] + GARCH.coef[2] * eps[m]^2 + GARCH.coef[3] * sigma.sq[m]
  quant.st     <- ( qst(p = alpha, xi=0, omega = 1, alpha = xi, nu) - E.wt ) / sqrt( nu / (nu - 2) - E.wt^2 )
  CVaR.true    <- ARMA.coef[1] * X[m] + sqrt(sigma.sq.np1) * quant.st  - ARMA.coef[2] * eps[m]
  
  # ... and CES
  quant.st     <- ( qst(p = alpha, xi=0, omega = 1, alpha = xi, nu) - E.wt ) / sqrt( nu / (nu - 2) - E.wt^2 )
  prux.st      <- integrate( sf.st, lower = -Inf, upper = quant.st, xi, nu )$value
  ES.st        <- quant.st - 1/alpha * prux.st
  CES.true     <- ARMA.coef[1] * X[m] + sqrt(sigma.sq.np1) * ES.st - ARMA.coef[2] * eps[m]
  
  return( list(X[(ntrans + 1) : m], CVaR.true, CES.true) )
}

# Calculate survivor function of standardized skewed-t distribution
sf.st <- function(x, xi, nu){
  E.wt  <- xi / sqrt(1 + xi^2) * sqrt(nu / pi) * gamma(0.5 * (nu - 1)) / gamma(0.5 * nu)
  sd.wt <- sqrt( nu / (nu - 2) - E.wt^2 )
  return( pst(sd.wt * x + E.wt, xi=0, omega=1, alpha=xi, nu) )
}

###########################

plot.coverage <- function(k, coverage.H.NA, coverage.H.SN, coverage.MR.NA, coverage.MR.SN, 
                          IL.H.NA, IL.H.SN, IL.MR.NA, IL.MR.SN, H.Bias, H.RMSE, MR.Bias, MR.RMSE){
  split.screen(figs=c(3, 2))
  
  screen(1)                            
  par(mar = c(4, 4.5, 2, 4.5))
  plot(k, coverage.H.NA, xlab="", ylab="" , ylim=c(0, 1), yaxs="i", type = "l", las=1)
  lines(k, coverage.H.SN, lty="dotted")
  abline(h = 0.95, lwd = 0.5)
  mtext(expression(italic(k)), side=1, line=2.5)
  mtext("Coverage probability", side = 2, line=3)
  
  screen(2)                            
  par(mar = c(4, 4.5, 2, 4.5))
  plot(k, coverage.MR.NA, xlab="", ylab="" , ylim=c(0, 1), yaxs="i", type = "l", las=1)
  lines(k, coverage.MR.SN, lty="dotted")
  abline(h = 0.95, lwd = 0.5)
  mtext(expression(italic(k)), side=1, line=2.5)
  mtext("Coverage probability", side = 2, line=3)
  
  screen(3)                            
  par(mar = c(5, 4.5, 1, 4.5))
  plot(k, IL.H.NA, xlab="", ylab="", ylim=c(0, 1.1 * max(c(IL.H.NA, IL.MR.NA, IL.H.SN, IL.MR.SN) )), yaxs="i", type = "l", las=1)
  lines(k, IL.H.SN, lty="dotted")
  mtext(expression(italic(k)), side=1, line=2.5)
  mtext("Interval length", side = 2, line=3)
  
  screen(4)                            
  par(mar = c(5, 4.5, 1, 4.5))
  plot(k, IL.MR.NA, xlab="", ylab="", ylim=c(0, 1.1 * max(c(IL.H.NA, IL.MR.NA, IL.H.SN, IL.MR.SN) )), yaxs="i", type = "l", las=1)
  lines(k, IL.MR.SN, lty="dotted")
  mtext(expression(italic(k)), side=1, line=2.5)
  mtext("Interval length", side = 2, line=3)
  
  screen(5)
  par(mar = c(5.5, 4.5, 0.5, 4.5))
  plot(k, H.Bias, type = "l", xlab="", ylab = "", ylim = c(min(c(MR.Bias, H.Bias)), max(c(MR.Bias, H.Bias))), las=1)
  mtext(expression(italic(k)), side=1, line=2.5)
  mtext("Bias", side = 2, line=3)
  par(new = T)
  plot(k, H.RMSE, ylim = c(0, 1.1*max(c(MR.RMSE, H.RMSE))), yaxs="i", type="l", lty=3, axes=F, xlab=NA, ylab=NA, cex=1.2, las=1)
  axis(side = 4, las=1)
  mtext(side = 4, line = 3, 'RMSE')
  legend("topright",
         legend=c("Bias", "RMSE"),
         lty=c(1,3), cex=0.8)
  mtext("(a) Hill estimator", side = 1, line=4)
  
  screen(6)
  par(mar = c(5.5, 4.5, 0.5, 4.5))
  plot(k, MR.Bias, type = "l", xlab="", ylab = "", ylim = c(min(c(MR.Bias, H.Bias)), max(c(MR.Bias, H.Bias))), las=1)
  mtext(expression(italic(k)), side=1, line=2.5)
  mtext("Bias", side = 2, line=3)
  par(new = T)
  plot(k, MR.RMSE, ylim = c(0, 1.1*max(c(MR.RMSE, H.RMSE))), yaxs="i", type="l", lty=3, axes=F, xlab=NA, ylab=NA, cex=1.2, las=1)
  axis(side = 4, las=1)
  mtext(side = 4, line = 3, 'RMSE')
  legend("topright",
         legend=c("Bias", "RMSE"),
         lty=c(1,3), cex=0.8)
  mtext("(b) MR estimator", side = 1, line=4)
  
  close.screen(all = TRUE)
}


#################################################################################
########################## Subsection 3.2: Application ##########################
#################################################################################


################################################
############### Table 3: Results ############### 
################################################

# Backtesting of VaR/ES models
library(zoo)    # for coredata()-command
library(fGarch) # for garchFit()-command



#####################################################################################
# DJIA # DJIA # DJIA # DJIA # DJIA # DJIA # DJIA # DJIA # DJIA # DJIA # DJIA # DJIA #
#####################################################################################

# 1. Prepare data

DJIA.data <- read.csv("DJI.csv")
DJIA      <- DJIA.data$Adj.Close        # DJIA prices from 1990 to 2016
Datum     <- DJIA.data$Date             # read in dates and ...
Datum     <- as.Date(Datum)             # ... convert to date format
lr.DJIA   <- diff(log(DJIA))            # Investigate log-returns
N         <- length(lr.DJIA)
n         <- 1000                       # length of rolling window


# 2. Forecast CVaR and CES

alpha <- 0.005

DJIA.CVaR.forecasts.cea07  <- matrix(NA, nrow = 1, ncol = (N-n))
DJIA.CVaR.forecasts.Hog.H  <- matrix(NA, nrow = 1, ncol = (N-n)) 
DJIA.CVaR.forecasts.Hog.MR <- matrix(NA, nrow = 1, ncol = (N-n)) 

DJIA.CES.forecasts.cea07  <- matrix(NA, nrow = 1, ncol = (N-n))
DJIA.CES.forecasts.Hog.H  <- matrix(NA, nrow = 1, ncol = (N-n)) 
DJIA.CES.forecasts.Hog.MR <- matrix(NA, nrow = 1, ncol = (N-n)) 


for(j in n : (N-1) ){
  forecast <-  VaR.forecast(lr.DJIA[(j-n+1) : j], alpha, nu = 10)
  
  DJIA.CVaR.forecasts.cea07[ , (j-n+1)]  <- forecast$VaR.cea07
  DJIA.CVaR.forecasts.Hog.H[ , (j-n+1)]  <- forecast$VaR.H.Hog17
  DJIA.CVaR.forecasts.Hog.MR[ , (j-n+1)] <- forecast$VaR.MR.Hog17
  
  DJIA.CES.forecasts.cea07[ , (j-n+1)]  <- forecast$ES.cea07
  DJIA.CES.forecasts.Hog.H[ , (j-n+1)]  <- forecast$ES.H.Hog17
  DJIA.CES.forecasts.Hog.MR[ , (j-n+1)] <- forecast$ES.MR.Hog17
}


# 3. CVaR Backtest using UC test of Kupiec (1995) and CC test of Berkowitz et al. (2011)

X <- lr.DJIA[(n+1):N]

BackTest <- Backtest.VaR(data = X, VaR = DJIA.CVaR.forecasts.cea07[1, ], alpha, Lags = 5)
cat("p-value for Cea07 with alpha = ", alpha, "\n")
print("Backtest of Kupiec(1995):")
print(BackTest$pval.Kup95)
print("Backtest of BCP11:")
print(BackTest$pval.BCP11)
print("#############################################")
print(" ")

BackTest <- Backtest.VaR(data = X, VaR = DJIA.CVaR.forecasts.Hog.H[1, ], alpha, Lags = 5)
cat("p-value for Hog.H with alpha = ", alpha, "\n")
print("Backtest of Kupiec(1995):")
print(BackTest$pval.Kup95)
print("Backtest of BCP11:")
print(BackTest$pval.BCP11)
print("#############################################")
print(" ")

BackTest <- Backtest.VaR(data = X, VaR = DJIA.CVaR.forecasts.Hog.MR[1, ], alpha, Lags = 5)
cat("p-value for Hog.MR with alpha = ", alpha,  "\n")
print("Backtest of Kupiec(1995):")
print(BackTest$pval.Kup95)
print("Backtest of BCP11:")
print(BackTest$pval.BCP11)
print("#############################################")
print(" ")



# 4. CVaR and CES forecast comparison using quantile score and the AL log score of Taylor (2017, Eqn. (19))


BackTest <- VaR.backtest(data = X, VaR = DJIA.CVaR.forecasts.cea07[1, ], alpha)
cat("Quantile Score for Cea07 with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = DJIA.CVaR.forecasts.cea07[1, ], ES = DJIA.CES.forecasts.cea07[1, ], alpha)
cat("Score for Cea07 with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")

BackTest <- VaR.backtest(data = X, VaR = DJIA.CVaR.forecasts.Hog.H[1, ], alpha)
cat("Quantile Score for Hog.H with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = DJIA.CVaR.forecasts.Hog.H[1, ], ES = DJIA.CES.forecasts.Hog.H[1, ], alpha)
cat("Score for Hog.H with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")

BackTest <- VaR.backtest(data = X, VaR = DJIA.CVaR.forecasts.Hog.MR[1, ], alpha)
cat("Quantile Score for Hog.MR with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = DJIA.CVaR.forecasts.Hog.MR[1, ], ES = DJIA.CES.forecasts.Hog.MR[1, ], alpha)
cat("Score for Hog.MR with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")



# 5. Diebold Mariano (1995) Test of equal predictive ability

l1.VaR <- loss.VaR(data = X, VaR = DJIA.CVaR.forecasts.Hog.H[1, ], alpha[1])
l2.VaR <- loss.VaR(data = X, VaR = DJIA.CVaR.forecasts.cea07[1, ],  alpha[1])

l1.ES <- loss.ES(data = X, VaR = DJIA.CVaR.forecasts.Hog.H[1, ], ES = DJIA.CES.forecasts.Hog.H[1, ], alpha[1])
l2.ES <- loss.ES(data = X, VaR = DJIA.CVaR.forecasts.cea07[1, ], ES = DJIA.CES.forecasts.cea07[1, ], alpha[1])

print("ES:")
dm95.test(l1.ES, l2.ES, "two.sided")
print("VaR:")
dm95.test(l1.VaR, l2.VaR, "two.sided")



###########################################################################################
# NASDAQ # NASDAQ # NASDAQ # NASDAQ # NASDAQ # NASDAQ # NASDAQ # NASDAQ # NASDAQ # NASDAQ #
###########################################################################################

# 1. Prepare data

NASDAQ.data <- read.csv("IXIC.csv")
NASDAQ      <- NASDAQ.data$Adj.Close  # NASDAQ prices from 1984 to 2016
Datum       <- NASDAQ.data$Date       # read in dates and ...
Datum       <- as.Date(Datum)         # ... convert to date format
lr.NASDAQ   <- diff(log(NASDAQ))      # Investigate log-returns
N           <- length(lr.NASDAQ)
n           <- 1000                   # length of rolling window

# 2. Forecast CVaR and CES
alpha <- 0.005

NASDAQ.CVaR.forecasts.cea07  <- matrix(NA, nrow = 1, ncol = (N-n))
NASDAQ.CVaR.forecasts.Hog.H  <- matrix(NA, nrow = 1, ncol = (N-n)) 
NASDAQ.CVaR.forecasts.Hog.MR <- matrix(NA, nrow = 1, ncol = (N-n)) 

NASDAQ.CES.forecasts.cea07  <- matrix(NA, nrow = 1, ncol = (N-n))
NASDAQ.CES.forecasts.Hog.H  <- matrix(NA, nrow = 1, ncol = (N-n)) 
NASDAQ.CES.forecasts.Hog.MR <- matrix(NA, nrow = 1, ncol = (N-n)) 


for(j in n : (N-1) ){
  forecast <-  VaR.forecast(lr.NASDAQ[(j-n+1) : j], alpha, nu = 10)
  
  NASDAQ.CVaR.forecasts.cea07[ , (j-n+1)]  <- forecast$VaR.cea07
  NASDAQ.CVaR.forecasts.Hog.H[ , (j-n+1)]  <- forecast$VaR.H.Hog17
  NASDAQ.CVaR.forecasts.Hog.MR[ , (j-n+1)] <- forecast$VaR.MR.Hog17
  
  NASDAQ.CES.forecasts.cea07[ , (j-n+1)]  <- forecast$ES.cea07
  NASDAQ.CES.forecasts.Hog.H[ , (j-n+1)]  <- forecast$ES.H.Hog17
  NASDAQ.CES.forecasts.Hog.MR[ , (j-n+1)] <- forecast$ES.MR.Hog17
}


# 3. CVaR Backtest using UC test of Kupiec (1995) and CC test of Berkowitz et al. (2011)

X <- lr.NASDAQ[(n+1):N]

BackTest <- Backtest.VaR(data = X, VaR = NASDAQ.CVaR.forecasts.cea07[1, ], alpha, Lags = 5)
cat("p-value for Cea07 with alpha = ", alpha, "\n")
print("Backtest of Kupiec(1995):")
print(BackTest$pval.Kup95)
print("Backtest of BCP11:")
print(BackTest$pval.BCP11)
print("#############################################")
print(" ")

BackTest <- Backtest.VaR(data = X, VaR = NASDAQ.CVaR.forecasts.Hog.H[1, ], alpha, Lags = 5)
cat("p-value for Hog.H with alpha = ", alpha, "\n")
print("Backtest of Kupiec(1995):")
print(BackTest$pval.Kup95)
print("Backtest of BCP11:")
print(BackTest$pval.BCP11)
print("#############################################")
print(" ")

BackTest <- Backtest.VaR(data = X, VaR = NASDAQ.CVaR.forecasts.Hog.MR[1, ], alpha, Lags = 5)
cat("p-value for Hog.MR with alpha = ", alpha, "\n")
print("Backtest of Kupiec(1995):")
print(BackTest$pval.Kup95)
print("Backtest of BCP11:")
print(BackTest$pval.BCP11)
print("#############################################")
print(" ")



# 4. CVaR and CES forecast comparison using quantile score and the AL log score of Taylor (2017, Eqn. (19))


BackTest <- VaR.backtest(data = X, VaR = NASDAQ.CVaR.forecasts.cea07[1, ], alpha)
cat("Quantile Score for Cea07 with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = NASDAQ.CVaR.forecasts.cea07[1, ], ES = NASDAQ.CES.forecasts.cea07[1, ], alpha)
cat("Score for Cea07 with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")

BackTest <- VaR.backtest(data = X, VaR = NASDAQ.CVaR.forecasts.Hog.H[1, ], alpha)
cat("Quantile Score for Hog.H with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = NASDAQ.CVaR.forecasts.Hog.H[1, ], ES = NASDAQ.CES.forecasts.Hog.H[1, ], alpha)
cat("Score for Hog.H with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")

BackTest <- VaR.backtest(data = X, VaR = NASDAQ.CVaR.forecasts.Hog.MR[1, ], alpha)
cat("Quantile Score for Hog.MR with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = NASDAQ.CVaR.forecasts.Hog.MR[1, ], ES = NASDAQ.CES.forecasts.Hog.MR[1, ], alpha)
cat("Score for Hog.MR with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")



# 5. Diebold Mariano (1995) Test of equal predictive ability

l1.VaR <- loss.VaR(data = X, VaR = NASDAQ.CVaR.forecasts.Hog.H[1, ], alpha[1])
l2.VaR <- loss.VaR(data = X, VaR = NASDAQ.CVaR.forecasts.cea07[1, ],  alpha[1])

l1.ES  <- loss.ES(data = X, VaR = NASDAQ.CVaR.forecasts.Hog.H[1, ], ES = NASDAQ.CES.forecasts.Hog.H[1, ], alpha[1])
l2.ES  <- loss.ES(data = X, VaR = NASDAQ.CVaR.forecasts.cea07[1, ],  ES = NASDAQ.CES.forecasts.cea07[1, ],  alpha[1])

print("ES:")
dm95.test(l1.ES, l2.ES, "two.sided")
print("VaR:")
dm95.test(l1.VaR, l2.VaR, "two.sided")



############################################################################################
# Nikkei 225 # Nikkei 225 # Nikkei 225 # Nikkei 225 # Nikkei 225 # Nikkei 225 # Nikkei 225 #
############################################################################################

# 1. Prepare data

N225.data <- read.csv("N225.csv")
N225      <- N225.data$Adj.Close        # N225 prices from 1984 to 2016
N225 <- as.numeric(as.character(N225))  # convert "factor" to "numeric"
N225 <- na.omit(N225)                   # delete NAs
Datum     <- N225.data$Date             # read in dates and ...
Datum     <- as.Date(Datum)             # ... convert to date format
lr.N225   <- diff(log(N225))            # Investigate log-returns
N         <- length(lr.N225)
n         <- 1000                       # length of rolling window


# 2. Forecast CVaR and CES

alpha <- 0.005

N225.CVaR.forecasts.cea07  <- matrix(NA, nrow = 1, ncol = (N-n))
N225.CVaR.forecasts.Hog.H  <- matrix(NA, nrow = 1, ncol = (N-n)) 
N225.CVaR.forecasts.Hog.MR <- matrix(NA, nrow = 1, ncol = (N-n)) 

N225.CES.forecasts.cea07  <- matrix(NA, nrow = 1, ncol = (N-n))
N225.CES.forecasts.Hog.H  <- matrix(NA, nrow = 1, ncol = (N-n)) 
N225.CES.forecasts.Hog.MR <- matrix(NA, nrow = 1, ncol = (N-n)) 


for(j in n : (N-1) ){
  forecast <-  VaR.forecast(lr.N225[(j-n+1) : j], alpha, nu = 10)
  
  N225.CVaR.forecasts.cea07[ , (j-n+1)]  <- forecast$VaR.cea07
  N225.CVaR.forecasts.Hog.H[ , (j-n+1)]  <- forecast$VaR.H.Hog17
  N225.CVaR.forecasts.Hog.MR[ , (j-n+1)] <- forecast$VaR.MR.Hog17
  
  N225.CES.forecasts.cea07[ , (j-n+1)]  <- forecast$ES.cea07
  N225.CES.forecasts.Hog.H[ , (j-n+1)]  <- forecast$ES.H.Hog17
  N225.CES.forecasts.Hog.MR[ , (j-n+1)] <- forecast$ES.MR.Hog17
}


# 3. CVaR Backtest using UC test of Kupiec (1995) and CC test of Berkowitz et al. (2011)

X <- lr.N225[(n+1):N]

BackTest <- Backtest.VaR(data = X, VaR = N225.CVaR.forecasts.cea07[1, ], alpha, Lags = 5)
cat("p-value for Cea07 with alpha = ", alpha, "\n")
print("Backtest of Kupiec(1995):")
print(BackTest$pval.Kup95)
print("Backtest of BCP11:")
print(BackTest$pval.BCP11)
print("#############################################")
print(" ")

BackTest <- Backtest.VaR(data = X, VaR = N225.CVaR.forecasts.Hog.H[1, ], alpha, Lags = 5)
cat("p-value for Hog.H with alpha = ", alpha, "\n")
print("Backtest of Kupiec(1995):")
print(BackTest$pval.Kup95)
print("Backtest of BCP11:")
print(BackTest$pval.BCP11)
print("#############################################")
print(" ")

BackTest <- Backtest.VaR(data = X, VaR = N225.CVaR.forecasts.Hog.MR[1, ], alpha, Lags = 5)
cat("p-value for Hog.MR with alpha = ", alpha, "\n")
print("Backtest of Kupiec(1995):")
print(BackTest$pval.Kup95)
print("Backtest of BCP11:")
print(BackTest$pval.BCP11)
print("#############################################")
print(" ")



# 4. CVaR and CES forecast comparison using quantile score and the AL log score of Taylor (2017, Eqn. (19))


BackTest <- VaR.backtest(data = X, VaR = N225.CVaR.forecasts.cea07[1, ], alpha)
cat("Quantile Score for Cea07 with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = N225.CVaR.forecasts.cea07[1, ], ES = N225.CES.forecasts.cea07[1, ], alpha)
cat("Score for Cea07 with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")

BackTest <- VaR.backtest(data = X, VaR = N225.CVaR.forecasts.Hog.H[1, ], alpha)
cat("Quantile Score for Hog.H with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = N225.CVaR.forecasts.Hog.H[1, ], ES = N225.CES.forecasts.Hog.H[1, ], alpha)
cat("Score for Hog.H with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")

BackTest <- VaR.backtest(data = X, VaR = N225.CVaR.forecasts.Hog.MR[1, ], alpha)
cat("Quantile Score for Hog.MR with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = N225.CVaR.forecasts.Hog.MR[1, ], ES = N225.CES.forecasts.Hog.MR[1, ], alpha)
cat("Score for Hog.MR with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")



# 5. Diebold Mariano (1995) Test of equal predictive ability

l1.VaR <- loss.VaR(data = X, VaR = N225.CVaR.forecasts.Hog.H[1, ], alpha[1])
l2.VaR <- loss.VaR(data = X, VaR = N225.CVaR.forecasts.cea07[1, ],  alpha[1])

l1.ES <- loss.ES(data = X, VaR = N225.CVaR.forecasts.Hog.H[1, ], ES = N225.CES.forecasts.Hog.H[1, ], alpha[1])
l2.ES <- loss.ES(data = X, VaR = N225.CVaR.forecasts.cea07[1, ], ES = N225.CES.forecasts.cea07[1, ], alpha[1])

print("ES:")
dm95.test(l1.ES, l2.ES, "two.sided")
print("VaR:")
dm95.test(l1.VaR, l2.VaR, "two.sided")



###########################################################################################
# HSI # HSI # HSI # HSI # HSI # HSI # HSI # HSI # HSI # HSI # HSI # HSI # HSI # HSI # HSI #
###########################################################################################

# 1. Prepare data

HSI.data  <- read.csv("HSI.csv")
HSI       <- HSI.data$Close        # HSI prices from 1997 to 2016
HSI       <- as.numeric(as.character(HSI))  # convert "factor" to "numeric"
HSI       <- HSI[!is.na(HSI)]
Datum     <- HSI.data$Date             # read in dates and ...
Datum     <- as.Date(Datum)             # ... convert to date format
lr.HSI    <- diff(log(HSI))            # Investigate log-returns
N         <- length(lr.HSI)
n         <- 1000                       # length of rolling window


# 2. Forecast CVaR and CES

alpha <- 0.005

HSI.CVaR.forecasts.cea07  <- matrix(NA, nrow = 1, ncol = (N-n))
HSI.CVaR.forecasts.Hog.H  <- matrix(NA, nrow = 1, ncol = (N-n)) 
HSI.CVaR.forecasts.Hog.MR <- matrix(NA, nrow = 1, ncol = (N-n)) 

HSI.CES.forecasts.cea07  <- matrix(NA, nrow = 1, ncol = (N-n))
HSI.CES.forecasts.Hog.H  <- matrix(NA, nrow = 1, ncol = (N-n)) 
HSI.CES.forecasts.Hog.MR <- matrix(NA, nrow = 1, ncol = (N-n)) 


for(j in n : (N-1) ){
  forecast <-  VaR.forecast(lr.HSI[(j-n+1) : j], alpha, nu = 10)
  
  HSI.CVaR.forecasts.cea07[ , (j-n+1)]  <- forecast$VaR.cea07
  HSI.CVaR.forecasts.Hog.H[ , (j-n+1)]  <- forecast$VaR.H.Hog17
  HSI.CVaR.forecasts.Hog.MR[ , (j-n+1)] <- forecast$VaR.MR.Hog17
  
  HSI.CES.forecasts.cea07[ , (j-n+1)]  <- forecast$ES.cea07
  HSI.CES.forecasts.Hog.H[ , (j-n+1)]  <- forecast$ES.H.Hog17
  HSI.CES.forecasts.Hog.MR[ , (j-n+1)] <- forecast$ES.MR.Hog17
}


# 3. CVaR Backtest using UC test of Kupiec (1995) and CC test of Berkowitz et al. (2011)

X <- lr.HSI[(n+1):N]

BackTest <- Backtest.VaR(data = X, VaR = HSI.CVaR.forecasts.cea07[1, ], alpha, Lags = 5)
cat("p-value for Cea07 with alpha = ", alpha, "\n")
print("Backtest of Kupiec(1995):")
print(BackTest$pval.Kup95)
print("Backtest of BCP11:")
print(BackTest$pval.BCP11)
print("#############################################")
print(" ")

BackTest <- Backtest.VaR(data = X, VaR = HSI.CVaR.forecasts.Hog.H[1, ], alpha, Lags = 5)
cat("p-value for Hog.H with alpha = ", alpha, "\n")
print("Backtest of Kupiec(1995):")
print(BackTest$pval.Kup95)
print("Backtest of BCP11:")
print(BackTest$pval.BCP11)
print("#############################################")
print(" ")

BackTest <- Backtest.VaR(data = X, VaR = HSI.CVaR.forecasts.Hog.MR[1, ], alpha, Lags =5)
cat("p-value for Hog.MR with alpha = ", alpha, "\n")
print("Backtest of Kupiec(1995):")
print(BackTest$pval.Kup95)
print("Backtest of BCP11:")
print(BackTest$pval.BCP11)
print("#############################################")
print(" ")



# 4. CVaR and CES forecast comparison using quantile score and the AL log score of Taylor (2017, Eqn. (19))


BackTest <- VaR.backtest(data = X, VaR = HSI.CVaR.forecasts.cea07[1, ], alpha)
cat("Quantile Score for Cea07 with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = HSI.CVaR.forecasts.cea07[1, ], ES = HSI.CES.forecasts.cea07[1, ], alpha)
cat("ES Score for Cea07 with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")

BackTest <- VaR.backtest(data = X, VaR = HSI.CVaR.forecasts.Hog.H[1, ], alpha)
cat("Quantile Score for Hog.H with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = HSI.CVaR.forecasts.Hog.H[1, ], ES = HSI.CES.forecasts.Hog.H[1, ], alpha)
cat("Score for Hog.H with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")

BackTest <- VaR.backtest(data = X, VaR = HSI.CVaR.forecasts.Hog.MR[1, ], alpha)
cat("Quantile Score for Hog.MR with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = HSI.CVaR.forecasts.Hog.MR[1, ], ES = HSI.CES.forecasts.Hog.MR[1, ], alpha)
cat("Score for Hog.MR with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")



# 5. Diebold Mariano (1995) Test of equal predictive ability

l1.VaR <- loss.VaR(data = X, VaR = HSI.CVaR.forecasts.Hog.H[1, ], alpha[1])
l2.VaR <- loss.VaR(data = X, VaR = HSI.CVaR.forecasts.cea07[1, ],  alpha[1])

l1.ES <- loss.ES(data = X, VaR = HSI.CVaR.forecasts.Hog.H[1, ], ES = HSI.CES.forecasts.Hog.H[1, ], alpha[1])
l2.ES <- loss.ES(data = X, VaR = HSI.CVaR.forecasts.cea07[1, ], ES = HSI.CES.forecasts.cea07[1, ], alpha[1])

print("ES:")
dm95.test(l1.ES, l2.ES, "two.sided")
print("VaR:")
dm95.test(l1.VaR, l2.VaR, "two.sided")



#########################################################################################
# CAC40 # CAC40 # CAC40 # CAC40 # CAC40 # CAC40 # CAC40 # CAC40 # CAC40 # CAC40 # CAC40 # 
#########################################################################################

# 1. Prepare data

CAC.data  <- read.csv("CAC40.csv")
CAC       <- CAC.data$Adj.Close        # CAC prices from 1990 to 2016
CAC       <- as.numeric(as.character(CAC))  # convert "factor" to "numeric"
CAC       <- CAC[!is.na(CAC)]
Datum     <- CAC.data$Date             # read in dates and ...
Datum     <- as.Date(Datum)             # ... convert to date format
lr.CAC    <- diff(log(CAC))            # Investigate log-returns
N         <- length(lr.CAC)
n         <- 1000                       # length of rolling window


# 2. Forecast CVaR and CES

alpha <- 0.005

CAC.CVaR.forecasts.cea07  <- matrix(NA, nrow = 1, ncol = (N-n))
CAC.CVaR.forecasts.Hog.H  <- matrix(NA, nrow = 1, ncol = (N-n)) 
CAC.CVaR.forecasts.Hog.MR <- matrix(NA, nrow = 1, ncol = (N-n)) 

CAC.CES.forecasts.cea07  <- matrix(NA, nrow = 1, ncol = (N-n))
CAC.CES.forecasts.Hog.H  <- matrix(NA, nrow = 1, ncol = (N-n)) 
CAC.CES.forecasts.Hog.MR <- matrix(NA, nrow = 1, ncol = (N-n)) 


for(j in n : (N-1) ){
  forecast <-  VaR.forecast(lr.CAC[(j-n+1) : j], alpha, nu = 10)
  
  CAC.CVaR.forecasts.cea07[ , (j-n+1)]  <- forecast$VaR.cea07
  CAC.CVaR.forecasts.Hog.H[ , (j-n+1)]  <- forecast$VaR.H.Hog17
  CAC.CVaR.forecasts.Hog.MR[ , (j-n+1)] <- forecast$VaR.MR.Hog17
  
  CAC.CES.forecasts.cea07[ , (j-n+1)]  <- forecast$ES.cea07
  CAC.CES.forecasts.Hog.H[ , (j-n+1)]  <- forecast$ES.H.Hog17
  CAC.CES.forecasts.Hog.MR[ , (j-n+1)] <- forecast$ES.MR.Hog17
}


# 3. CVaR Backtest using UC test of Kupiec (1995) and CC test of Berkowitz et al. (2011)

X <- lr.CAC[(n+1):N]
for(i in 1:1){
  BackTest <- Backtest.VaR(data = X, VaR = CAC.CVaR.forecasts.cea07[1, ], alpha, Lags = 5)
  cat("p-value for Cea07 with alpha = ", alpha, "\n")
  print("Backtest of Kupiec(1995):")
  print(BackTest$pval.Kup95)
  print("Backtest of BCP11:")
  print(BackTest$pval.BCP11)
  print("#############################################")
  print(" ")
  
  BackTest <- Backtest.VaR(data = X, VaR = CAC.CVaR.forecasts.Hog.H[1, ], alpha, Lags = 5)
  cat("p-value for Hog.H with alpha = ", alpha, "\n")
  print("Backtest of Kupiec(1995):")
  print(BackTest$pval.Kup95)
  print("Backtest of BCP11:")
  print(BackTest$pval.BCP11)
  print("#############################################")
  print(" ")
  
  BackTest <- Backtest.VaR(data = X, VaR = CAC.CVaR.forecasts.Hog.MR[1, ], alpha, Lags = 5)
  cat("p-value for Hog.MR with alpha = ", alpha, "\n")
  print("Backtest of Kupiec(1995):")
  print(BackTest$pval.Kup95)
  print("Backtest of BCP11:")
  print(BackTest$pval.BCP11)
  print("#############################################")
  print(" ")
}


# 4. CVaR and CES forecast comparison using quantile score and the AL log score of Taylor (2017, Eqn. (19))


BackTest <- VaR.backtest(data = X, VaR = CAC.CVaR.forecasts.cea07[1, ], alpha)
cat("Quantile Score for Cea07 with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = CAC.CVaR.forecasts.cea07[1, ], ES = CAC.CES.forecasts.cea07[1, ], alpha)
cat("Score for Cea07 with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")

BackTest <- VaR.backtest(data = X, VaR = CAC.CVaR.forecasts.Hog.H[1, ], alpha)
cat("Quantile Score for Hog.H with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = CAC.CVaR.forecasts.Hog.H[1, ], ES = CAC.CES.forecasts.Hog.H[1, ], alpha)
cat("Score for Hog.H with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")

BackTest <- VaR.backtest(data = X, VaR = CAC.CVaR.forecasts.Hog.MR[1, ], alpha)
cat("Quantile Score for Hog.MR with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = CAC.CVaR.forecasts.Hog.MR[1, ], ES = CAC.CES.forecasts.Hog.MR[1, ], alpha)
cat("Score for Hog.MR with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")



# 5. Diebold Mariano (1995) Test of equal predictive ability

l1.VaR <- loss.VaR(data = X, VaR = CAC.CVaR.forecasts.Hog.H[1, ], alpha[1])
l2.VaR <- loss.VaR(data = X, VaR = CAC.CVaR.forecasts.cea07[1, ],  alpha[1])

l1.ES <- loss.ES(data = X, VaR = CAC.CVaR.forecasts.Hog.H[1, ], ES = CAC.CES.forecasts.Hog.H[1, ], alpha[1])
l2.ES <- loss.ES(data = X, VaR = CAC.CVaR.forecasts.cea07[1, ], ES = CAC.CES.forecasts.cea07[1, ], alpha[1])

print("ES:")
dm95.test(l1.ES, l2.ES, "two.sided")
print("VaR:")
dm95.test(l1.VaR, l2.VaR, "two.sided")



#########################################################################################
# DAX30 # DAX30 # DAX30 # DAX30 # DAX30 # DAX30 # DAX30 # DAX30 # DAX30 # DAX30 # DAX30 #
#########################################################################################

# 1. Prepare data

DAX.data  <- read.csv("GDAXI.csv")
DAX       <- DAX.data$Adj.Close             # DAX prices from 1990 to 2016
DAX       <- as.numeric(as.character(DAX))  # convert "factor" to "numeric"
DAX       <- DAX[!is.na(DAX)]
Datum     <- DAX.data$Date                  # read in dates and ...
Datum     <- as.Date(Datum)                 # ... convert to date format
lr.DAX    <- diff(log(DAX))                 # Investigate log-returns
N         <- length(lr.DAX)
n         <- 1000                           # length of rolling window


# 2. Forecast CVaR and CES

alpha <- 0.005

DAX.CVaR.forecasts.cea07  <- matrix(NA, nrow = 1, ncol = (N-n))
DAX.CVaR.forecasts.Hog.H  <- matrix(NA, nrow = 1, ncol = (N-n)) 
DAX.CVaR.forecasts.Hog.MR <- matrix(NA, nrow = 1, ncol = (N-n)) 

DAX.CES.forecasts.cea07  <- matrix(NA, nrow = 1, ncol = (N-n))
DAX.CES.forecasts.Hog.H  <- matrix(NA, nrow = 1, ncol = (N-n)) 
DAX.CES.forecasts.Hog.MR <- matrix(NA, nrow = 1, ncol = (N-n)) 


for(j in n : (N-1) ){
  forecast <-  VaR.forecast(lr.DAX[(j-n+1) : j], alpha, nu = 10)
  
  DAX.CVaR.forecasts.cea07[ , (j-n+1)]  <- forecast$VaR.cea07
  DAX.CVaR.forecasts.Hog.H[ , (j-n+1)]  <- forecast$VaR.H.Hog17
  DAX.CVaR.forecasts.Hog.MR[ , (j-n+1)] <- forecast$VaR.MR.Hog17
  
  DAX.CES.forecasts.cea07[ , (j-n+1)]  <- forecast$ES.cea07
  DAX.CES.forecasts.Hog.H[ , (j-n+1)]  <- forecast$ES.H.Hog17
  DAX.CES.forecasts.Hog.MR[ , (j-n+1)] <- forecast$ES.MR.Hog17
}


# 3. CVaR Backtest using UC test of Kupiec (1995) and CC test of Berkowitz et al. (2011)

X <- lr.DAX[(n+1):N]

BackTest <- Backtest.VaR(data = X, VaR = DAX.CVaR.forecasts.cea07[1, ], alpha, Lags = 5)
cat("p-value for Cea07 with alpha = ", alpha, "\n")
print("Backtest of Kupiec(1995):")
print(BackTest$pval.Kup95)
print("Backtest of BCP11:")
print(BackTest$pval.BCP11)
print("#############################################")
print(" ")

BackTest <- Backtest.VaR(data = X, VaR = DAX.CVaR.forecasts.Hog.H[1, ], alpha, Lags = 5)
cat("p-value for Hog.H with alpha = ", alpha, "\n")
print("Backtest of Kupiec(1995):")
print(BackTest$pval.Kup95)
print("Backtest of BCP11:")
print(BackTest$pval.BCP11)
print("#############################################")
print(" ")

BackTest <- Backtest.VaR(data = X, VaR = DAX.CVaR.forecasts.Hog.MR[1, ], alpha, Lags = 5)
cat("p-value for Hog.MR with alpha = ", alpha, "\n")
print("Backtest of Kupiec(1995):")
print(BackTest$pval.Kup95)
print("Backtest of BCP11:")
print(BackTest$pval.BCP11)
print("#############################################")
print(" ")


# 4. CVaR and CES forecast comparison using quantile score and the AL log score of Taylor (2017, Eqn. (19))

BackTest <- VaR.backtest(data = X, VaR = DAX.CVaR.forecasts.cea07[1, ], alpha)
cat("Quantile Score for Cea07 with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = DAX.CVaR.forecasts.cea07[1, ], ES = DAX.CES.forecasts.cea07[1, ], alpha)
cat("Score for Cea07 with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")

BackTest <- VaR.backtest(data = X, VaR = DAX.CVaR.forecasts.Hog.H[1, ], alpha)
cat("Quantile Score for Hog.H with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = DAX.CVaR.forecasts.Hog.H[1, ], ES = DAX.CES.forecasts.Hog.H[1, ], alpha)
cat("Score for Hog.H with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")

BackTest <- VaR.backtest(data = X, VaR = DAX.CVaR.forecasts.Hog.MR[1, ], alpha)
cat("Quantile Score for Hog.MR with alpha = ", alpha, ": ",BackTest, "\n")
BackTest <- ES.backtest(data = X, VaR = DAX.CVaR.forecasts.Hog.MR[1, ], ES = DAX.CES.forecasts.Hog.MR[1, ], alpha)
cat("Score for Hog.MR with alpha = ", alpha, ": ",BackTest, "\n")
print("#############################################")
print(" ")



# 5. Diebold Mariano (1995) Test of equal predictive ability

l1.VaR <- loss.VaR(data = X, VaR = DAX.CVaR.forecasts.Hog.H[1, ], alpha)
l2.VaR <- loss.VaR(data = X, VaR = DAX.CVaR.forecasts.cea07[1, ],  alpha)

l1.ES <- loss.ES(data = X, VaR = DAX.CVaR.forecasts.Hog.H[1, ], ES = DAX.CES.forecasts.Hog.H[1, ], alpha)
l2.ES <- loss.ES(data = X, VaR = DAX.CVaR.forecasts.cea07[1, ], ES = DAX.CES.forecasts.cea07[1, ], alpha)

print("ES:")
dm95.test(l1.ES, l2.ES, "two.sided")
print("VaR:")
dm95.test(l1.VaR, l2.VaR, "two.sided")



##############################################################
############### Functions to replicate Table 3 ############### 
##############################################################


# Function to generate VaR forecasts

VaR.forecast <- function(X, alpha, nu){  # VaR forecasts using Chan et al. (2007) and Hoga (2017+)
  n <- length(X)
  m <- n - nu + 1
  no.alpha <- length(alpha)
  
  #####################################
  
  # 1. [Cea07]
  # 1.1. Estimate GARCH(1,1) model using standard QMLE for log-returns
  param <- garchFit(formula = ~garch(1,1), data = X, init.rec = "mci", cond.dist = "QMLE", 
                    include.mean = FALSE,  include.delta = FALSE, include.skew = FALSE, 
                    include.shape = FALSE, algorithm = "lbfgsb", trace = FALSE)@fit$matcoef
  theta.hat <- as.numeric( param[ , 1] )
  theta.hat <- c(0, 0, theta.hat[1:3])   # save the GARCH(1,1) parameters
  
  # 1.2. Compute residuals U.hat
  eps.hat      <- c(X[1], coredata(filter(X[2:n]-theta.hat[1]*X[1:(n-1)],
                                          filter = theta.hat[2], method = "recursive", init = X[1])))
  sigma.sq.hat <- c(theta.hat[3], coredata(filter(theta.hat[3] + theta.hat[4] * eps.hat[1:(n-1)]^2,
                                                  filter = theta.hat[5], method = "recursive", init = theta.hat[3])))
  
  U.hat <- -eps.hat / sqrt(sigma.sq.hat)   # [KL16, p. 243]
  U.hat <- c(rep.int(NA, nu-1), U.hat[nu:length(U.hat)])       # offset effects of intialization of variance equation
  
  # 1.3.1 Compute optimal k* using Fixed Choice of [Cea07]
  k.star.H  <- floor(1.5 * log(n)^2)    
  
  # 1.3.2 Calculate EQ and ES estimates using k.star
  
  H.EQ <- numeric(no.alpha)
  sort.U <- sort.int(U.hat, decreasing = TRUE, method = "quick")
  H    <- 1/(k.star.H) * sum(log(sort.U[1 : (floor(k.star.H)+1)] / sort.U[floor(k.star.H)+1]))  # Hill estimator
  for(p in 1 : no.alpha){
    H.EQ[p] <- -sort.U[floor(k.star.H)+1] * ((m*alpha[p]) / k.star.H)^(-H)                # H EQ estimator
  }
  H.ES <- H.EQ / (1 - H)
  
  sigma.hat.n1   <- sqrt( theta.hat[3] + theta.hat[4] * eps.hat[n]^2 + theta.hat[5] * sigma.sq.hat[n] )
  CVaR.est.Cea07 <- theta.hat[1] * X[n] + sigma.hat.n1 * H.EQ  - theta.hat[2] * eps.hat[n]   # [HogXX, Eq. (1)]
  CES.est.Cea07  <- theta.hat[1] * X[n] + sigma.hat.n1 * H.ES  - theta.hat[2] * eps.hat[n]   # [HogXX, Eq. (1)]
  
  #####################################
  
  # 2. [Hog17+]
  
  # 2.1. Estimate AR(1)-GARCH(1,1) model using standard QMLE for log-returns
  param <- garchFit(formula = ~arma(1,0)+garch(1,1), data = X, init.rec = "mci", cond.dist = "QMLE",
                    include.mean = FALSE,  include.delta = FALSE, include.skew = FALSE,
                    include.shape = FALSE, algorithm = "lbfgsb", trace = FALSE)@fit$matcoef
  theta.hat <- as.numeric(param[ ,1])
  theta.hat <- append(theta.hat, 0, 1)
  
  
  # 2.2. Compute residuals U.hat
  eps.hat <- c(X[1], coredata(filter(X[2:n]-theta.hat[1]*X[1:(n-1)],
                                     filter = theta.hat[2], method = "recursive", init = X[1])))
  sigma.sq.hat <- c(theta.hat[3], coredata(filter(theta.hat[3] + theta.hat[4] * eps.hat[1:(n-1)]^2,
                                                  filter = theta.hat[5], method = "recursive", init = theta.hat[3])))
  
  U.hat <- -eps.hat / sqrt(sigma.sq.hat)   # [KL16, p. 243]
  U.hat <- c(rep.int(NA, nu-1), U.hat[nu:length(U.hat)])       # offset effects of intialization of variance equation
  sort.U <- sort.int(U.hat, decreasing = TRUE, method = "quick")
  
  # 2.3.1 Compute optimal k* using [Dea16+, p. 16] (Tail Index Estimation: Quantile Driven Threshold Selection)
  
  k.min <- 50
  k.max <- 200
  
  sup.H <- sup.MR <- rep.int(NA, k.max)
  for(k in k.min : k.max){
    H.k    <- 1/k * sum(log( sort.U[1 : (k+1)] / sort.U[(k+1)] ))
    MR.k   <- 0.5 * ( 1/k * sum(log( sort.U[1 : (k+1)] / sort.U[(k+1)] )^2) ) / H.k
    sup.H[k] <- max( abs(sort.U[2 : (k.max+1)] - (k/(1 : k.max))^( H.k ) * sort.U[(k+1)]) )
    sup.MR[k]<- max( abs(sort.U[2 : (k.max+1)] - (k/(1 : k.max))^( MR.k ) * sort.U[(k+1)]) )
  }
  k.star.H  <- which.min(sup.H)
  k.star.MR <- which.min(sup.MR)
  
  # 2.3.2 Calculate sequential EQ estimates using k.star
  
  # H estimator
  
  H.EQ <- numeric(no.alpha)
  H    <- 1/(k.star.H) * sum(log(sort.U[1 : (floor(k.star.H)+1)] / sort.U[floor(k.star.H)+1])) # Hill estimator
  for(p in 1 : no.alpha){
    H.EQ[p] <- -sort.U[floor(k.star.H)+1] * ((n*alpha[p]) / k.star.H)^(-H)   # H EQ estimator
  }
  H.ES <- H.EQ / (1 - H)
  
  # MR estimator
  
  MR.EQ <- numeric(no.alpha)
  M1    <- 1/(k.star.MR) * sum(log(sort.U[1 : (floor(k.star.MR)+1)] / sort.U[floor(k.star.MR)+1])) # Hill estimator
  M2    <- 1/(k.star.MR) * sum(log(sort.U[1 : (floor(k.star.MR)+1)] / sort.U[floor(k.star.MR) + 1])^2)
  MR    <- 0.5 * M2/M1                                                    # MR estimator
  for(p in 1:no.alpha){
    MR.EQ[p] <- -sort.U[floor(k.star.MR)+1] * ((n*alpha[p]) / k.star.MR)^(-MR)   # MR EQ estimator
  }
  MR.ES <- MR.EQ / (1 - MR)
  
  sigma.hat.n1 <- sqrt( theta.hat[3] + theta.hat[4] * eps.hat[n]^2 + theta.hat[5] * sigma.sq.hat[n] )
  
  CVaR.est.H.Hog17 <- theta.hat[1] * X[n] + sigma.hat.n1 * H.EQ  - theta.hat[2] * eps.hat[n]   # [HogXX, Eq. (1)]
  CES.est.H.Hog17  <- theta.hat[1] * X[n] + sigma.hat.n1 * H.ES  - theta.hat[2] * eps.hat[n]   # [HogXX, Eq. (1)]
  
  CVaR.est.MR.Hog17 <- theta.hat[1] * X[n] + sigma.hat.n1 * MR.EQ - theta.hat[2] * eps.hat[n]   # [HogXX, Eq. (1)]
  CES.est.MR.Hog17  <- theta.hat[1] * X[n] + sigma.hat.n1 * MR.ES - theta.hat[2] * eps.hat[n]   # [HogXX, Eq. (1)]
  

  return( list( VaR.cea07 = c(CVaR.est.Cea07), VaR.H.Hog17 = c(CVaR.est.H.Hog17), 
                VaR.MR.Hog17 = c(CVaR.est.MR.Hog17), ES.cea07 = c(CES.est.Cea07), 
                ES.H.Hog17 = c(CES.est.H.Hog17), ES.MR.Hog17 = c(CES.est.MR.Hog17) ) )
}

#######################

Backtest.VaR <- function(data, VaR, alpha, Lags = 5){
  # Apply basic version of tests of Escanciano and Olmo (2010)
  
  # Kupiec's (1995) UC test
  P       <- length(data)
  Ind.hit <- 1*(data <= VaR)
  Kup95   <- 1/sqrt(alpha * (1-alpha)) * 1/sqrt(P) * sum(Ind.hit - alpha)
  p.Kup95 <- 2 * pnorm( -abs(Kup95) )
  
  # Berkowitz, Christoffersen and Pelletier's (2011) CC test
  gamma.Pj <- numeric(Lags)
  mu.hit   <- Ind.hit - alpha
  for(j in 1:Lags){
    gamma.Pj[j] <- 1 / (P - j) * sum( mu.hit[(j+1):P] * mu.hit[1:(P-j)] )
  }
  LB      <- P * (P+2) * sum( (P - 1:Lags)^(-1) * ( gamma.Pj / (alpha * (1 - alpha)) )^2 )
  p.BCP11 <- 1 - pchisq(LB, Lags)
  
  return(list(pval.Kup95 = p.Kup95, pval.BCP11 = p.BCP11))
}


#######################

# VaR Backtest using the quantile score of Taylor (2017, Eqn. (14))

VaR.backtest <- function(data, VaR, alpha){
  S <- sum( (data - VaR) * (alpha - 1*(data <= VaR)) )
  return(S)
}

#######################

# ES Backtest using the AL log score of Taylor (2017, Eqn. (19))

ES.backtest <- function(data, VaR, ES, alpha){
  S <- sum( -log((alpha-1) / ES) - (data - VaR) * (alpha - 1*(data <= VaR)) / (alpha * ES) )
  return(S)
}

#######################

# ES loss via AL log score

loss.ES <- function(data, VaR, ES, alpha){
  S <- -log((alpha-1) / ES) - (data - VaR) * (alpha - 1*(data <= VaR)) / (alpha * ES)
  return(S)
}

#######################

# VaR loss via tick loss function

loss.VaR <- function(data, VaR, alpha){
  S <- (data - VaR) * (alpha - 1*(data <= VaR)) 
  return(S)
}

#######################

# Diebold-Mariano (1995) test of equal predicitve ability

dm95.test <- function (loss1, loss2, alternative = c("two.sided", "less", "greater"), h = 1){
  alternative <- match.arg(alternative)
  d <- c(loss1) - c(loss2)
  d.cov <- acf(d, na.action = na.omit, lag.max = h - 1, type = "covariance",
               plot = FALSE)$acf[, , 1]
  d.var <- sum(c(d.cov[1], 2 * d.cov[-1]))/length(d)
  dv <- d.var
  if (dv > 0)
    STATISTIC <- mean(d, na.rm = TRUE)/sqrt(dv)
  else if (h == 1)
    stop("Variance of DM statistic is zero")
  else {
    warning("Variance is negative, using horizon h=1")
    return(dm95.test(loss1, loss2, alternative, h = 1))
  }
  n <- length(d)
  k <- ((n + 1 - 2 * h + (h/n) * (h - 1))/n)^(1/2)
  STATISTIC <- STATISTIC * k
  names(STATISTIC) <- "DM"
  if (alternative == "two.sided")
    PVAL <- 2 * pt(-abs(STATISTIC), df = n - 1)
  else if (alternative == "less")
    PVAL <- pt(STATISTIC, df = n - 1)
  else if (alternative == "greater")
    PVAL <- pt(STATISTIC, df = n - 1, lower.tail = FALSE)
  return(PVAL)
}

################################################################################